/*
 * File: Xml.i
 *
 * XML module specific to Lua
 */

%{
#include "terralib/Version.h"

struct ProjectMetadata
{
  std::string file;										    //!< The project file.
  std::string title;                                       //!< The title of the project.
  std::string author;                                     //!< The author of the project.
  std::vector<te::map::AbstractLayer*> layers;

  std::vector<te::map::AbstractLayer*> getLayers()
  {
    return layers;
  }
};

void SaveProject(const std::string& file, const std::string& author, const std::string& title,
                 const std::vector<te::map::AbstractLayer*>& layers)
{
  std::unique_ptr<te::xml::AbstractWriter> writer(te::xml::AbstractWriterFactory::make());

  writer->setURI(file);

  std::string schema_loc = te::core::FindInTerraLibPath("share/terralib/schemas/terralib/qt/af/project.xsd");

  writer->writeStartDocument("UTF-8", "no");

  writer->writeStartElement("Project");

  boost::replace_all(schema_loc, " ", "%20");

  writer->writeAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema-instance");
  writer->writeAttribute("xmlns:te_da", "http://www.terralib.org/schemas/dataaccess");
  writer->writeAttribute("xmlns:te_map", "http://www.terralib.org/schemas/maptools");
  writer->writeAttribute("xmlns:te_qt_af", "http://www.terralib.org/schemas/common/af");

  writer->writeAttribute("xmlns:se", "http://www.opengis.net/se");
  writer->writeAttribute("xmlns:ogc", "http://www.opengis.net/ogc");
  writer->writeAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");

  writer->writeAttribute("xmlns", "http://www.terralib.org/schemas/qt/af");
  writer->writeAttribute("xsd:schemaLocation", "http://www.terralib.org/schemas/qt/af " + schema_loc);
  writer->writeAttribute("version", TERRALIB_VERSION_STRING);

  writer->writeElement("Title", title);
  writer->writeElement("Author", author);

  //write data source list
  writer->writeStartElement("te_da:DataSourceList");

  writer->writeAttribute("xmlns:te_common", "http://www.terralib.org/schemas/common");

  te::da::DataSourceInfoManager::iterator itBegin = te::da::DataSourceInfoManager::getInstance().begin();
  te::da::DataSourceInfoManager::iterator itEnd = te::da::DataSourceInfoManager::getInstance().end();
  te::da::DataSourceInfoManager::iterator it;

  std::vector<std::string> dsIdVec;
  for (size_t i = 0; i < layers.size(); i++)
  {
    dsIdVec.push_back(layers[i]->getDataSourceId());
  }

  for(it=itBegin; it!=itEnd; ++it)
  {

    if (std::find(dsIdVec.begin(), dsIdVec.end(), it->second->getId()) == dsIdVec.end())
      continue;

    writer->writeStartElement("te_da:DataSource");

    writer->writeAttribute("id", it->second->getId());
    writer->writeAttribute("type", it->second->getType());
    writer->writeAttribute("access_driver", it->second->getAccessDriver());

    writer->writeStartElement("te_da:Title");
    writer->writeValue(it->second->getTitle());
    writer->writeEndElement("te_da:Title");

    writer->writeStartElement("te_da:Description");
    writer->writeValue(it->second->getDescription());
    writer->writeEndElement("te_da:Description");

    writer->writeStartElement("te_da:URI");
	std::string connInfo(te::core::URIEncode(it->second->getConnInfoAsString()));
	boost::replace_all(connInfo, "&", "%26");
    writer->writeValue(connInfo);
    writer->writeEndElement("te_da:URI");

    writer->writeEndElement("te_da:DataSource");
  }

  writer->writeEndElement("te_da:DataSourceList");
  //end write

  writer->writeStartElement("ComponentList");
  writer->writeEndElement("ComponentList");

  writer->writeStartElement("te_map:LayerList");

  const te::map::serialize::Layer& lserial = te::map::serialize::Layer::getInstance();

  for (size_t i = 0; i < layers.size(); i++)
  {
    lserial.write(layers[i], *writer.get());
  }

  writer->writeEndElement("te_map:LayerList");

  writer->writeEndElement("Project");

  writer->writeToFile();
}

ProjectMetadata LoadProject(const std::string& projFile)
{
  ProjectMetadata proj;
  std::unique_ptr<te::xml::Reader> xmlReader(te::xml::ReaderFactory::make());

  xmlReader->read(projFile);

  if(!xmlReader->next())
  {
    std::string msg = "Could not read project information in the file: " + projFile + ".";
    throw te::common::Exception(msg);
  }

  if(xmlReader->getNodeType() != te::xml::START_ELEMENT)
  {
    std::string msg = "Error reading the document " + projFile + ", the start element wasn't found.";
    throw te::common::Exception(msg);
  }

  if(xmlReader->getElementLocalName() != "Project")
  {
    std::string msg = "The first tag in the document " + projFile + " is not 'Project'.";
    throw te::common::Exception(msg);
  }

  xmlReader->next();
  assert(xmlReader->getNodeType() == te::xml::START_ELEMENT);
  assert(xmlReader->getElementLocalName() == "Title");

  xmlReader->next();
  assert(xmlReader->getNodeType() == te::xml::VALUE);
  proj.title = xmlReader->getElementValue().c_str();

  xmlReader->next(); // End element

  xmlReader->next();
  assert(xmlReader->getNodeType() == te::xml::START_ELEMENT);
  assert(xmlReader->getElementLocalName() == "Author");

  xmlReader->next();

  if(xmlReader->getNodeType() == te::xml::VALUE)
  {
    proj.author = xmlReader->getElementValue().c_str();
    xmlReader->next(); // End element
  }

  //  //read data source list from this project
  xmlReader->next();

  assert(xmlReader->getNodeType() == te::xml::START_ELEMENT);
  assert(xmlReader->getElementLocalName() == "DataSourceList");

  xmlReader->next();

  // DataSourceList contract form
  if(xmlReader->getNodeType() == te::xml::END_ELEMENT &&
     xmlReader->getElementLocalName() == "DataSourceList")
  {
    xmlReader->next();
  }

  while((xmlReader->getNodeType() == te::xml::START_ELEMENT) &&
        (xmlReader->getElementLocalName() == "DataSource"))
  {
    te::da::DataSourceInfoPtr ds(te::serialize::xml::ReadDataSourceInfo(*(xmlReader.get())));
    te::da::DataSourceInfoManager::getInstance().add(ds);
  }

  //  //end read data source list

  assert(xmlReader->getNodeType() == te::xml::START_ELEMENT);
  assert(xmlReader->getElementLocalName() == "ComponentList");
  xmlReader->next(); // End element
  xmlReader->next(); // next after </ComponentList>

  assert(xmlReader->getNodeType() == te::xml::START_ELEMENT);
  assert(xmlReader->getElementLocalName() == "LayerList");

  xmlReader->next();

  const te::map::serialize::Layer& lserial = te::map::serialize::Layer::getInstance();

  // Read the layers
  while((xmlReader->getNodeType() != te::xml::END_ELEMENT) &&
        (xmlReader->getElementLocalName() != "LayerList"))
  {
    te::map::AbstractLayer* layer(lserial.read(*(xmlReader.get())));
    assert(layer);
    proj.layers.push_back(layer);
  }

  assert(xmlReader->getNodeType() == te::xml::END_ELEMENT);
  assert(xmlReader->getElementLocalName() == "LayerList");

  xmlReader->next();
  assert((xmlReader->getNodeType() == te::xml::END_ELEMENT) || (xmlReader->getNodeType() == te::xml::END_DOCUMENT));
  assert(xmlReader->getElementLocalName() == "Project");

  proj.file = projFile;

  return proj;
}

class QgisLayerInfo
{
public:
  void setUri(const te::core::URI& uri)
  {
    this->uri = uri;
  }

  te::core::URI getUri()
  {
    return uri;
  }

  void setName(const std::string& name)
  {
    this->name = name;
  }

  std::string getName()
  {
	  return name;
  }

  void setSrid(int srid)
  {
	  this->srid = srid;
  }

  int getSrid()
  {
	  return srid;
  }

  void setProj4(const std::string& proj4)
  {
	  this->proj4 = proj4;
  }

  std::string getProj4()
  {
	  return proj4;
  }

private:
  std::string name;
  int srid;
  te::core::URI uri;
  std::string proj4;
};

class QgisProjectInfo
{
public:
  void setFile(const std::string& path)
  {
    this->file = path;
  }

  std::string getFile()
  {
    return file;
  }

  void setTitle(const std::string& title)
  {
    this->title = title;
  }

  std::string getTitle()
  {
    return title;
  }

  void setAuthor(const std::string& author)
  {
    this->author = author;
  }

  std::string getAuthor()
  {
    return author;
  }

  void addLayer(QgisLayerInfo* layer)
  {
	  layers.push_back(layer);
  }

  std::vector<QgisLayerInfo*> getLayers()
  {
    return layers;
  }

private:
  std::string file;
  std::string title;
  std::string author;
  std::vector<QgisLayerInfo*> layers;
};

#include <boost/filesystem.hpp>

class QgisReader
{
public:
  QgisReader(xercesc::DOMElement* element)
  {
    this->element = element;
  }

  te::core::URI getElementUriByName(const char* name, const boost::filesystem::path& relativeTo)
  {
	xercesc::DOMNodeList* node = element->getElementsByTagName(xercesc::XMLString::transcode(name));
	xercesc::DOMElement* el = dynamic_cast<xercesc::DOMElement*>(node->item(0));

	std::string content(xercesc::XMLString::transcode(el->getTextContent()));

	if (isDatabaseContent(content))
    {
	  return createDatabaseUri(content);
    }
	else if (isWfsContent(content))
    {
	  return createWfsUri(content);
    }
	else if (isWmsContent(content))
    {
      return createWmsUri(content);
    }
    else
    {
      try
      {
	    boost::filesystem::path dir(relativeTo.parent_path());
	    boost::filesystem::path canonic(boost::filesystem::canonical(content, dir));
	    return te::core::URI("file://" + canonic.string());
      }
	  catch (const boost::filesystem::filesystem_error& e)
      {
        throw std::runtime_error(content + " - " + e.what());
      }
    }
    
    return te::core::URI("");
  }

  std::string getElementStringByName(const char* name)
  {
	  xercesc::DOMNodeList* node = element->getElementsByTagName(xercesc::XMLString::transcode(name));
	  xercesc::DOMElement* el = dynamic_cast<xercesc::DOMElement*>(node->item(0));
	  std::string value = xercesc::XMLString::transcode(el->getTextContent());

	  return value;
  }

private:
  std::string removeQuotes(const std::string& str)
  {
	  if ((str.find("\'") != std::string::npos) || (str.find("\"") != std::string::npos))
	{
	  return str.substr(1, str.size() - 2);
    }

    return str;
  }

  std::pair<std::string, std::string> split(const std::string& str, const std::string& token)
  {
    size_t pos = str.find(token);
	if (pos != std::string::npos)
    {
	  std::string key(str.substr(0, pos));
	  std::string value(str.substr(pos + 1));
	  return std::pair<std::string, std::string>(key, value);
    }

	return std::pair<std::string, std::string>(str, str);
  }

  std::string formatTable(const std::string& str)
  {
    std::pair<std::string, std::string> result(split(str, "."));

	return removeQuotes(result.second);
  }

  std::string getAttribute(xercesc::DOMElement* element, const char* attr)
  {
	  return xercesc::XMLString::transcode(element->getAttribute(xercesc::XMLString::transcode(attr)));
  }

  bool hasAttribute(const std::string& content, const std::string& attr)
  {
	  return content.find(attr) != std::string::npos;
  }

  bool isDatabaseContent(const std::string& content)
  {
    return hasAttribute(content, "dbname");
  }

  bool isWfsContent(const std::string& content)
  {
    return hasAttribute(content, "typename");
  }

  bool isWmsContent(const std::string& content)
  {
	  return hasAttribute(content, "contextualWMSLegend");
  }

  std::map<std::string, std::string> createMapAttributeValues(const std::string& content, const char* separator)
  {
    boost::char_separator<char> sep(separator);
	boost::tokenizer<boost::char_separator<char>> tokens(content, sep);
	std::map<std::string, std::string> contents;

	for (boost::tokenizer< boost::char_separator<char> >::iterator it = tokens.begin(); it != tokens.end(); it++)
	{
	  std::string token(*it);

	  if (token.find("=") != std::string::npos)
	  {
	    contents.insert(split(token, "="));
	  }
	}

    return contents;
  }

  te::core::URI createDatabaseUri(const std::string& content)
  {
    std::map<std::string, std::string> contents(createMapAttributeValues(content, " "));

	std::string uriStr("pgsql://" + removeQuotes(contents.at("user")) + ":"
	       	                   + removeQuotes(contents.at("password")) + "@"
			                   + removeQuotes(contents.at("host")) + ":"
			                   + removeQuotes(contents.at("port")) + "/"
			                   + removeQuotes(contents.at("dbname")) + "?"
			                   + formatTable(contents.at("table")));

	te::core::URI uri(uriStr);


    if (!uri.isValid())
	    throw(std::runtime_error("Invalid QGIS database URI."));
      
    return uri;
  }

  te::core::URI createWfsUri(const std::string& content)
  {
    std::map<std::string, std::string> contents(createMapAttributeValues(content, " "));

	std::string uriStr("wfs:" + removeQuotes(contents.at("url")) + "?"
                              + removeQuotes(contents.at("typename")));

    te::core::URI uri(uriStr);

	  if (!uri.isValid())
	    throw(std::runtime_error("Invalid QGIS WFS URI."));

		return uri;
  }

  te::core::URI createWmsUri(const std::string& content)
  {
    std::map<std::string, std::string> contents(createMapAttributeValues(content, "&"));

    std::pair<std::string, std::string> format(split(contents.at("format"), "/"));

	std::string uriStr("wms:" + removeQuotes(contents.at("url")) + "?"
	               	                + "format=" + format.second + "&"
                                    + "layers=" + removeQuotes(contents.at("layers")));

    te::core::URI uri(uriStr);

	if (!uri.isValid())
	  throw(std::runtime_error("Invalid QGIS WMS URI."));
	  
	return uri;

  }

  xercesc::DOMElement* element;
};

class QgisProject
{
public:
  static QgisProjectInfo load(const std::string& projFile)
  {
    QgisProjectInfo proj;

    try
    {
      xercesc::DOMDocument* doc = te::xerces::readXML(projFile);
      xercesc::DOMElement* root = doc->getDocumentElement();

    if (!root)
      throw(std::runtime_error("empty QGIS project."));

    xercesc::DOMNodeList* children = root->getChildNodes();

	for (unsigned int i = 0; i < children->getLength(); i++)
    {
	  xercesc::DOMNode* currentNode = children->item(i);

	  if (currentNode->getNodeType() &&  // true is not NULL
		  currentNode->getNodeType() == xercesc::DOMNode::ELEMENT_NODE) // is element
	  {
		// Found node which is an Element. Re-cast node as element
		xercesc::DOMElement* currentElement = dynamic_cast<xercesc::DOMElement*>(currentNode);

		if (xercesc::XMLString::equals(currentElement->getTagName(), xercesc::XMLString::transcode("title")))
		{
		  std::string title = xercesc::XMLString::transcode(currentElement->getTextContent());
		  proj.setTitle(title);
		}
		else if (xercesc::XMLString::equals(currentElement->getTagName(), xercesc::XMLString::transcode("projectlayers")))
		{
		  xercesc::DOMNodeList* mlNode = currentElement->getElementsByTagName(xercesc::XMLString::transcode("maplayer"));

		  for (unsigned int j = 0; j < mlNode->getLength(); j++)
		  {
			xercesc::DOMNode* currNode = mlNode->item(j);

			if (currNode->getNodeType() &&  // true is not NULL
				currNode->getNodeType() == xercesc::DOMNode::ELEMENT_NODE) // is element
			{
			  xercesc::DOMElement* currEl = dynamic_cast<xercesc::DOMElement*>(currNode);
			  QgisReader reader(currEl);
			  QgisLayerInfo* layer(new QgisLayerInfo());

			  std::string srid(reader.getElementStringByName("srid"));
			  layer->setSrid(std::stoi(srid));

			  std::string name(reader.getElementStringByName("layername"));
			  layer->setName(name);

			  try
              {
		  	    te::core::URI uri = reader.getElementUriByName("datasource", projFile);
				layer->setUri(uri);

				proj.addLayer(layer);
              }
              catch (const std::runtime_error& e)
              {
                std::cout << "Warning: Layer '" << name << "' was ignored. "
                            << "File not found '" << e.what() << "'."
                            << std::endl;
              }
            }
		  }
		}
	  }
    }
  }
  catch (xercesc::XMLException& e)
  {
	char* message = xercesc::XMLString::transcode(e.getMessage());
	xercesc::XMLString::release(&message);
  }

  return proj;
}
};

%}

struct ProjectMetadata
{
  std::string file;										    //!< The project file.
  std::string title;                                       //!< The title of the project.
  std::string author;                                     //!< The author of the project.

  std::vector<te::map::AbstractLayer*> getLayers();
};

void SaveProject(const std::string& file, const std::string& author, const std::string& title,
                 const std::vector<te::map::AbstractLayer*>& layers);

ProjectMetadata LoadProject(const std::string& projFile);

class QgisLayerInfo
{
public:
  void setUri(const te::core::URI& uri);
  te::core::URI getUri();
  void setName(const std::string& name);
  std::string getName();
  void setSrid(int srid);
  int getSrid();
  void setProj4(const std::string& proj4);
  std::string getProj4();
};

class QgisProjectInfo
{
public:
  void setFile(const std::string& path);
  std::string getFile();
  void setTitle(const std::string& title);
  std::string getTitle();
  void setAuthor(const std::string& author);
  std::string getAuthor();
  void addLayer(QgisLayerInfo* layer);
  std::vector<QgisLayerInfo*> getLayers();
};

class QgisProject
{
public:
  static QgisProjectInfo load(const std::string& projFile);
};
