/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/tools/Info.cpp

  \brief This class implements a concrete tool to get informations about a specified feature using pointing operation.
*/

// TerraLib
#include "../../../common/STLUtils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/Coord2D.h"
#include "../../../geometry/Envelope.h"
#include "../../../geometry/Geometry.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../geometry/Point.h"
#include "../../../geometry/Utils.h"
#include "../../../maptools/MarkRendererManager.h"
#include "../../../raster/Raster.h"
#include "../../../raster/RasterProperty.h"
#include "../../../raster/Utils.h"
#include "../../../se/Fill.h"
#include "../../../se/Stroke.h"
#include "../../../se/Mark.h"
#include "../../../se/Utils.h"
#include "../../../srs/Config.h"
#include "../canvas/Canvas.h"
#include "../canvas/MapDisplay.h"
#include "../Utils.h"
#include "../layer/info/FeatureInfoMakerFactory.h"
#include "../layer/info/AbstractFeatureInfoMaker.h"
#include "Info.h"

// Qt
#include <QtCore/QPointF>
#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QMessageBox>
#include <QMouseEvent>

// STL
#include <cassert>
#include <memory>

te::qt::widgets::Info::Info(te::qt::widgets::MapDisplay* display, const QCursor& cursor, const std::list<te::map::AbstractLayerPtr>& layers, QObject* parent) 
  : AbstractTool(display, parent),
    m_layers(layers),
    m_infoWidget(new QTreeWidget(display))
{
  setCursor(cursor);

  // Setup the widget that will be used to show the informations
  m_infoWidget->setWindowTitle(tr("Information"));
  m_infoWidget->setWindowFlags(Qt::Tool);
  m_infoWidget->setAlternatingRowColors(true);
  m_infoWidget->setMinimumSize(300, 300);
  m_infoWidget->setColumnCount(2);

  QStringList labels;
  labels << tr("Property") << tr("Value");
  m_infoWidget->setHeaderLabels(labels);
}

te::qt::widgets::Info::~Info()
{
  QPixmap* draft = m_display->getDraftPixmap();

  draft->fill(Qt::transparent);

  m_infoWidget->close();

  delete m_infoWidget;
}

bool te::qt::widgets::Info::mouseReleaseEvent(QMouseEvent* e)
{
  if(e->button() != Qt::LeftButton)
    return false;

  Qt::KeyboardModifiers keys = e->modifiers();

  if(keys & Qt::ShiftModifier)
    m_layers = m_display->getLayerList();

  if(m_layers.empty())
    return false;

  QPointF pixelOffset(4.0, 4.0);
#if (QT_VERSION >= 0x050000)
  QRectF rect = QRectF(e->localPos() - pixelOffset, e->localPos() + pixelOffset);
#else
  QRectF rect = QRectF(e->posF() - pixelOffset, e->posF() + pixelOffset);
#endif
      
  // Converts rect boundary to world coordinates
  QPointF ll(rect.left(), rect.bottom());
  QPointF ur(rect.right(), rect.top());
  ll = m_display->transform(ll);
  ur = m_display->transform(ur);

  // Bulding the query box
  te::gm::Envelope envelope(ll.x(), ll.y(), ur.x(), ur.y());

  // Clear draft!
  QPixmap* draft = m_display->getDraftPixmap();
  draft->fill(Qt::transparent);

  // Clear info widget!
  m_infoWidget->clear();

  // For each layer, get info!
  std::list<te::map::AbstractLayerPtr>::const_iterator it;
  for(it = m_layers.begin(); it != m_layers.end(); ++it)
  {
    te::map::AbstractLayerPtr layer = *it;

    // The Layer item
    QTreeWidgetItem* layerItem = new QTreeWidgetItem;
    layerItem->setIcon(0, QIcon::fromTheme("layer"));
    layerItem->setText(0, tr("Layer"));
    layerItem->setText(1, layer->getTitle().c_str());

    m_infoWidget->addTopLevelItem(layerItem);

    std::unique_ptr<AbstractFeatureInfoMaker> featureInfoMaker;

    featureInfoMaker.reset(te::qt::widgets::FeatureInfoMakerFactory::make(layer->getType()));

    featureInfoMaker->getInfo(layerItem, m_display, layer, envelope);

    if(layerItem->childCount() == 0)
      delete layerItem;
  }

  if(m_infoWidget->topLevelItemCount() > 0)
  {
    m_infoWidget->expandAll();
    m_infoWidget->resizeColumnToContents(0);
    m_infoWidget->show();
  }
  else
    m_infoWidget->hide();

  m_display->repaint();

  return true;
}

void te::qt::widgets::Info::setLayers(const std::list<te::map::AbstractLayerPtr>& layers)
{
  m_layers = layers;
}
