/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/utils/FileDialog.cpp

  \brief An instance of QFileDialog that includes file extention if necessary.
*/

// TerraLib
#include "../../../common/StringUtils.h"
#include "../../../core/translator/Translator.h"
#include "../../../dataaccess/datasource/DataSource.h"
#include "../../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../../dataaccess/datasource/DataSourceFactory.h"
#include "../Utils.h"
#include "FileDialog.h"

// STL
#include <map>
#include <memory>
#include <string>

// Qt
#include <QString>

// BOOST
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>

te::qt::widgets::FileDialog::FileDialog(QWidget *parent, FileType type, std::string fileInfo):
  m_parent(parent),
  m_type(type),
  m_fileInfo(fileInfo)
{
}

void te::qt::widgets::FileDialog::exec()
{
  QString extensions = "";
  std::map<QString, QString> mappedExtensions;

  if(m_type == RASTER)
    extensions = GetDiskRasterFileSelFilter(true);
  else if(m_type == VECTOR)
    extensions = GetDiskVectorFileSelFilter();

  mappedExtensions = getMappedExtensions(extensions);

  QString filteredExtention;

  QFileDialog fileDialog(m_parent);

  QString filePath = fileDialog.getSaveFileName(m_parent,
                                                QString(TE_TR("Save as...")),
                                                GetFilePathFromSettings(m_fileInfo.c_str()),
                                                extensions,
                                                &filteredExtention,
                                                QFileDialog::DontConfirmOverwrite);

  if (filePath.isEmpty())
    throw te::common::Exception(TE_TR("File path is empty"));

  // Verifies if filename has special characters.
  boost::filesystem::path path(filePath.toUtf8().data());
  std::string filename = path.stem().string();

  if(!path.extension().empty())
     if(mappedExtensions[filteredExtention].compare(path.extension().string().c_str()))
        throw te::common::Exception(TE_TR("The file extension is invalid."));

  if (hasInvalidCharacters(filename))
    throw te::common::Exception(TE_TR("File name has special characters."));

  // Append sufix in file path.
  QFileInfo fileInfo( filePath );
  if(fileInfo.suffix().isEmpty())
    filePath.append(mappedExtensions[filteredExtention]);

  // Set file information in class members.
  setPath(filePath.toUtf8().data());
  setFileName(filePath.toUtf8().data());
  setExtension(filePath.toUtf8().data());
}

std::string te::qt::widgets::FileDialog::getPath()
{
  return m_path;
}

void te::qt::widgets::FileDialog::setPath(std::string filePath)
{
  m_path = filePath;
}

std::string te::qt::widgets::FileDialog::getFileName()
{
  return m_fileName;
}

void te::qt::widgets::FileDialog::setFileName(std::string filePath)
{
  boost::filesystem::path path(filePath);
  m_fileName = path.leaf().string();
}

std::string te::qt::widgets::FileDialog::getExtension()
{
  return m_extension;
}

void te::qt::widgets::FileDialog::setExtension(std::string filePath)
{
  boost::filesystem::path path(filePath);
  m_extension = path.extension().string();
}

std::map<QString, QString> te::qt::widgets::FileDialog::getMappedExtensions(QString extensions)
{
  std::map<QString, QString> mappedExtensions;

  std::vector< std::string > tokens;
  te::common::Tokenize( extensions.toStdString(), tokens, ";;" );
  for(auto extension : tokens)
  {
    boost::regex expression("[.]+[a-z0-9]+");
    boost::smatch extMatched;
    if(boost::regex_search(extension, extMatched, expression))
    {
      mappedExtensions[QString::fromStdString(extension)] = QString::fromStdString(extMatched[0]);
    }
  }

  return mappedExtensions;
}

bool te::qt::widgets::FileDialog::hasInvalidCharacters(std::string filename)
{
  boost::regex expression{"\\W+"};

  if (boost::regex_search(filename, expression))
    return true;

  return false;
}

