/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/Functions.cpp

  \brief Raster Processing functions.
*/

// TerraLib

#ifdef WIN32
#define NOMINMAX
#endif

#include "Functions.h"

#include "../common/SystemApplicationSettings.h"
#include "../common/Version.h"
#include "../core/utils/Platform.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/datasource/DataSourceFactory.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/Enums.h"
#include "../raster/BandProperty.h"
#include "../raster/Grid.h"
#include "../raster/RasterFactory.h"
#include "../raster/RasterProperty.h"
#include "../raster/RasterIterator.h"
#include "../raster/Utils.h"
#include "../geometry/Point.h"
#include "../common/MatrixUtils.h"
#include "../common/progress/TaskProgress.h"
#include "Exception.h"
#include "Macros.h"
#include "RasterHandler.h"
#include "Contrast.h"

// Boost
#include <boost/filesystem.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/thread.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <boost/numeric/ublas/matrix.hpp>

// STL
#include <cstring>
#include <string>
#include <limits>
#include <map>
#include <memory>
#include <algorithm>

#ifndef M_PI
  #define M_PI       3.14159265358979323846
#endif

namespace te
{
  namespace rp
  {
    struct GetMeanValueThreadParams
    {
      te::rst::Band const * m_inputBandPtr;
      Matrix< bool >* m_rasterBlocksStatusPtr;
      double* m_meanValuePtr;
      double* m_pixelsNumberValuePtr;
      bool m_returnStatus;
      boost::mutex* m_mutexPtr;
    };  
    
    struct GetStdDevValueThreadParams
    {
      te::rst::Band const * m_inputBandPtr;
      Matrix< bool >* m_rasterBlocksStatusPtr;
      double* m_stdDevValuePtr;
      double* m_pixelsNumberValuePtr;
      double m_meanValue;
      bool m_returnStatus;
      boost::mutex* m_mutexPtr;
    };     
    
    struct GetCovarianceValueThreadParams
    {
      te::rst::Band const * m_inputBand1Ptr;
      te::rst::Band const * m_inputBand2Ptr;
      Matrix< bool >* m_rasterBlocksStatusPtr;
      double* m_covarianceValuePtr;
      double* m_pixelsNumberValuePtr;
      bool m_returnStatus;
      double m_mean1Value;
      double m_mean2Value;
      boost::mutex* m_mutexPtr;
    };   
    
    struct RemapValuesThreadParams
    {
      te::rst::Raster const * m_inputRasterPtr;
      std::vector< unsigned int > const * m_inputRasterBandsPtr;
      te::rst::Raster* m_outputRasterPtr;
      std::vector< unsigned int > const * m_outputRasterBandsPtr;
      boost::numeric::ublas::matrix< double > const * m_remapMatrixPtr;
      Matrix< bool > * m_rasterBlocksStatusPtr;
      bool m_returnStatus;
      boost::mutex* m_mutexPtr;
    };     
   
    bool CreateNewRaster( const te::rst::Grid& rasterGrid,
      const std::vector< te::rst::BandProperty* >& bandsProperties,
      const std::string& outDataSetName,
      const std::string& dataSourceType,
      RasterHandler& outRasterHandler )
    {
      // Creating a new datasource

      std::unique_ptr< te::da::DataSource > dataSourcePtr(
        te::da::DataSourceFactory::make( dataSourceType, ("file://" + outDataSetName)));
      if( dataSourcePtr.get() == 0 ) return false;      
      
      RasterHandler internalRasterHandler;
      
      if( CreateNewRaster( rasterGrid, bandsProperties, 
        outDataSetName, *dataSourcePtr, internalRasterHandler ) )
      {
        std::unique_ptr< te::da::DataSource > dummyDataSourcePtr;
        std::unique_ptr< te::da::DataSourceTransactor > transactorPtr;
        std::unique_ptr< te::da::DataSet > dataSetPtr;
        std::unique_ptr< te::rst::Raster > rasterPtr;
        
        internalRasterHandler.release( dummyDataSourcePtr, transactorPtr,
          dataSetPtr, rasterPtr );
          
        outRasterHandler.reset( dataSourcePtr.release(), transactorPtr.release(), 
          dataSetPtr.release(), rasterPtr.release() );
        
        return true;
      }
      else
      {
        return false;
      }      
    }
        
    bool CreateNewRaster( const te::rst::Grid& rasterGrid,
      const std::vector< te::rst::BandProperty* >& bandsProperties,
      const std::string& outDataSetName,
      te::da::DataSource& outDataSource,
      RasterHandler& outRasterHandler )
    {
      // Defining the raster properties

      std::unique_ptr< te::rst::RasterProperty > rasterPropertyPtr(
        new te::rst::RasterProperty( new te::rst::Grid( rasterGrid ),
        bandsProperties, std::map< std::string, std::string >(),
        false, 0, 0 ) );

      // acquiring a transactor instance

      std::unique_ptr< te::da::DataSourceTransactor > transactorPtr(
       outDataSource.getTransactor() );

      if( transactorPtr.get() == 0 )
      {
       return false;
      }

      // Creating a data set instance
      
      if( transactorPtr->dataSetExists( outDataSetName ) )
      {
        transactorPtr->dropDataSet( outDataSetName );
      }

      std::unique_ptr< te::da::DataSetType > dataSetTypePtr(
        new te::da::DataSetType( outDataSetName ) );
      dataSetTypePtr->add( rasterPropertyPtr.release() );

      try
      {
        transactorPtr->createDataSet( dataSetTypePtr.get(),
          std::map< std::string, std::string >() );
      }
      catch( ... )
      {
        return false;
      }
      
       if( ! transactorPtr->dataSetExists( outDataSetName ) ) return false;

      std::unique_ptr< te::da::DataSet > dataSetPtr( transactorPtr->getDataSet(
        outDataSetName, te::common::FORWARDONLY, true, te::common::RWAccess ) );

      if( dataSetPtr.get() == 0 )
      {
        return false;
      }

      // Creating a raster instance

      std::unique_ptr< te::rst::Raster > rasterPtr( dataSetPtr->getRaster( 0 ) );

      if( rasterPtr.get() )
      {
        outRasterHandler.reset( transactorPtr.release(), dataSetPtr.release(), rasterPtr.release() );
        return true;
      }
      else
      {
        return false;
      }
    }
    
    bool TERPEXPORT CreateNewRaster( const te::rst::Grid& rasterGrid,
      const std::vector< te::rst::BandProperty* >& bandsProperties,
      const std::map< std::string, std::string>& rasterInfo,
      const std::string& rasterType,
      std::unique_ptr< te::rst::Raster >& outRasterPtr )
    {
      try
      {
        outRasterPtr.reset( te::rst::RasterFactory::make( rasterType,
          new te::rst::Grid( rasterGrid ), bandsProperties, rasterInfo, 0, 0 ) );
      }
      catch( const te::common::Exception&/* exc */)
      {
        return false;
      }
      
      if( outRasterPtr.get() == 0 )
      {
        return false;
      }      
      
      return true;
    }

    bool CreateNewMemRaster( const te::rst::Grid& rasterGrid,
      std::vector< te::rst::BandProperty* > bandsProperties,
      RasterHandler& outRasterHandler )
    {
      std::string dataSetName = std::string( "createNewMemRaster" ) +
        boost::lexical_cast< std::string >( &outRasterHandler );
        
      return CreateNewRaster( rasterGrid, bandsProperties, dataSetName, 
        "MEM", outRasterHandler );
    }

    bool CreateNewGdalRaster( const te::rst::Grid& rasterGrid,
      std::vector< te::rst::BandProperty* > bandsProperties,
      const std::string& fileName,
      RasterHandler& outRasterHandler )
    {
      std::unique_ptr< te::rst::Raster > outRasterPtr;
      if( CreateNewGdalRaster( rasterGrid, bandsProperties, fileName, outRasterPtr ) )
      {
        outRasterHandler.reset( outRasterPtr.release() ); 
        return true;
      }
      else
      {
        outRasterHandler.reset();
        return false;
      }
    }
    
    bool CreateNewGdalRaster( const te::rst::Grid& rasterGrid,
      std::vector< te::rst::BandProperty* > bandsProperties,
      const std::string& fileName,
      std::unique_ptr< te::rst::Raster >& outRasterPtr )
    {
      outRasterPtr.reset();
      
      std::map< std::string, std::string > rInfo;
      rInfo[ "URI" ] = fileName;
      
      outRasterPtr.reset( te::rst::RasterFactory::make(
        "GDAL",
        new te::rst::Grid( rasterGrid ),
        bandsProperties,
        rInfo,
        0,
        0 ) );
      
      if( outRasterPtr.get() )
      {
        return true;
      }
      else
      {
        return false;
      }
    }
    
    bool TERPEXPORT Copy2DiskRaster( const te::rst::Raster& inputRaster,
      const std::string& fileName )
    {
      if( !(inputRaster.getAccessPolicy() & te::common::RAccess ) )
      {
        return false;
      };
      
      const unsigned int nBands = static_cast<unsigned int>(inputRaster.getNumberOfBands());
      const unsigned int nCols = inputRaster.getNumberOfColumns();
      const unsigned int nRows = inputRaster.getNumberOfRows();
      unsigned int bandIdx =0;
      unsigned int col = 0;
      unsigned int row = 0;
      
      std::vector< te::rst::BandProperty* > bandsProperties;
      for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
      {
        bandsProperties.push_back( new te::rst::BandProperty(
          *( inputRaster.getBand( bandIdx )->getProperty() ) ) );        
      }
      
      RasterHandler outRasterHandler;
      
      if( ! CreateNewGdalRaster( *( inputRaster.getGrid() ), bandsProperties,
        fileName, outRasterHandler ) )
      {
        return false;
      }
      
      double value = 0;
      
      for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
      {
        const te::rst::Band& inBand = *inputRaster.getBand( bandIdx );
        te::rst::Band& outBand = *outRasterHandler.getRasterPtr()->getBand( bandIdx );
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, value );
            outBand.setValue( col, row, value );
          }
        }
      }
      
      return true;
    }

    void Convert2DoublesVector( void* inputVector, const int inputVectorDataType,
      unsigned int inputVectorSize, double* outputVector )
    {
      switch( inputVectorDataType )
      {
        case te::dt::CHAR_TYPE :
        {
          char* vPtr = static_cast<char*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::BIT_TYPE :
        case te::dt::UCHAR_TYPE :
        {
          unsigned char* vPtr = static_cast<unsigned char*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::INT16_TYPE :
        {
          short int* vPtr = static_cast<short int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::CINT16_TYPE :
        {
          std::complex< short int >* vPtr = static_cast<std::complex< short int >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        case te::dt::UINT16_TYPE :
        {
          unsigned short int* vPtr = static_cast<unsigned short int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::INT32_TYPE :
        {
          int* vPtr = static_cast<int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::CINT32_TYPE :
        {
          std::complex< int >* vPtr = static_cast<std::complex< int >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        case te::dt::UINT32_TYPE :
        {
          unsigned int* vPtr = static_cast<unsigned int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::INT64_TYPE :
        {
          long int* vPtr = static_cast<long int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::UINT64_TYPE :
        {
          unsigned long int* vPtr = static_cast<unsigned long int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::FLOAT_TYPE :
        {
          float* vPtr = static_cast<float*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::CFLOAT_TYPE :
        {
          std::complex< float >* vPtr = static_cast<std::complex< float >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        case te::dt::DOUBLE_TYPE :
        {
          memcpy( outputVector, inputVector, inputVectorSize * sizeof( double ) );
          break;
        }
        case te::dt::CDOUBLE_TYPE :
        {
          std::complex< double >* vPtr = static_cast<std::complex< double >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        default :
          throw te::rp::Exception( "Invalid data type" );
      };
    }

    void ConvertDoublesVector( double* inputVector,
      unsigned int inputVectorSize, const int outputVectorDataType,
      void* outputVector )
    {
      switch( outputVectorDataType )
      {
        case te::dt::CHAR_TYPE :
        {
          char* vPtr = static_cast<char*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<char>(inputVector[ idx ]);
          break;
        }
        case te::dt::BIT_TYPE :
        case te::dt::UCHAR_TYPE :
        {
          unsigned char* vPtr = static_cast<unsigned char*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned char>(inputVector[ idx ]);
          break;
        }
        case te::dt::INT16_TYPE :
        {
          short int* vPtr = static_cast<short int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<short int>(inputVector[ idx ]);
          break;
        }
        case te::dt::CINT16_TYPE :
        {
          std::complex< short int >* vPtr = static_cast<std::complex< short int >*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ]= ( static_cast<short int>(inputVector[ idx ]) );
          break;
        }
        case te::dt::UINT16_TYPE :
        {
          unsigned short int* vPtr = static_cast<unsigned short int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned short int>(inputVector[ idx ]);
          break;
        }
        case te::dt::INT32_TYPE :
        {
          int* vPtr = static_cast<int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<int>(inputVector[ idx ]);
          break;
        }
        case te::dt::CINT32_TYPE :
        {
          std::complex< int >* vPtr = static_cast<std::complex< int >*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = ( static_cast<int>(inputVector[ idx ]) );
          break;
        }
        case te::dt::UINT32_TYPE :
        {
          unsigned int* vPtr = static_cast<unsigned int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned int>(inputVector[ idx ]);
          break;
        }
        case te::dt::INT64_TYPE :
        {
          long int* vPtr = static_cast<long int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<long int>(inputVector[ idx ]);
          break;
        }
        case te::dt::UINT64_TYPE :
        {
          unsigned long int* vPtr = static_cast<unsigned long int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned long int>(inputVector[ idx ]);
          break;
        }
        case te::dt::FLOAT_TYPE :
        {
          float* vPtr = static_cast<float*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<float>(inputVector[ idx ]);
          break;
        }
        case te::dt::CFLOAT_TYPE :
        {
          std::complex< float >* vPtr = static_cast<std::complex< float >*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = ( static_cast<float>(inputVector[ idx ]) );
          break;
        }
        case te::dt::DOUBLE_TYPE :
        {
          memcpy( outputVector, inputVector, inputVectorSize * sizeof( double ) );
          break;
        }
        case te::dt::CDOUBLE_TYPE :
        {
          std::complex< double >* vPtr = static_cast<std::complex< double >*>(outputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = ( static_cast<double>(inputVector[ idx ]) );
          break;
        }
        default :
          throw te::rp::Exception( "Invalid data type" );
      };
    }

    std::string GetSensorFilename()
    {
      std::string datalocalPath = te::core::GetAppLocalDataLocation();
      std::string app = te::common::SystemApplicationSettings::getInstance().getValue("Application.Name");
      std::string Organization = te::common::SystemApplicationSettings::getInstance().getValue("Application.Organization");
      std::string vers = te::common::Version::asString();

      static std::string sens_file = datalocalPath + "/" + Organization + "/" + app + "-" + vers + "/SpectralSensor.json";

      return sens_file;
    }
    
    std::vector<std::string> GetBandNames()
    {
      static std::vector<std::string> bandNames;
      bandNames.clear();

      bandNames.push_back("CBERS2_CCD_1_BLUE");
      bandNames.push_back("CBERS2_CCD_2_GREEN");
      bandNames.push_back("CBERS2_CCD_3_RED");
      bandNames.push_back("CBERS2_CCD_4_NIR");
      bandNames.push_back("CBERS2_CCD_5_PAN");

      bandNames.push_back("LANDSAT5_TM_1_BLUE");
      bandNames.push_back("LANDSAT5_TM_2_GREEN");
      bandNames.push_back("LANDSAT5_TM_3_RED");
      bandNames.push_back("LANDSAT5_TM_4_NIR");
      bandNames.push_back("LANDSAT5_TM_5_SWIR");
      bandNames.push_back("LANDSAT5_TM_6_THERMAL");
      bandNames.push_back("LANDSAT5_TM_7_MIR");

      bandNames.push_back("LANDSAT7_ETM+_1_BLUE");
      bandNames.push_back("LANDSAT7_ETM+_2_GREEN");
      bandNames.push_back("LANDSAT7_ETM+_3_RED");
      bandNames.push_back("LANDSAT7_ETM+_4_NIR");
      bandNames.push_back("LANDSAT7_ETM+_5_SWIR");
      bandNames.push_back("LANDSAT7_ETM+_6_THERMAL");
      bandNames.push_back("LANDSAT7_ETM+_7_MIR");
      bandNames.push_back("LANDSAT7_ETM+_8_PAN");

      bandNames.push_back("WV2_MUL_1_COASTAL");
      bandNames.push_back("WV2_MUL_2_BLUE");
      bandNames.push_back("WV2_MUL_3_GREEN");
      bandNames.push_back("WV2_MUL_4_YELLOW");
      bandNames.push_back("WV2_MUL_5_RED");
      bandNames.push_back("WV2_MUL_6_REDEDGE");
      bandNames.push_back("WV2_MUL_7_NIR1");
      bandNames.push_back("WV2_MUL_8_NIR2");
      bandNames.push_back("WV2_PAN");

      bandNames.push_back("AWIFS_RED");
      bandNames.push_back("AWIFS_GREEN");
      bandNames.push_back("AWIFS_NIR");
      bandNames.push_back("AWIFS_MIR");

      bandNames.push_back("RAPID_EYE_BLUE");
      bandNames.push_back("RAPID_EYE_GREEN");
      bandNames.push_back("RAPID_EYE_RED");
      bandNames.push_back("RAPID_EYE_REDEDGE");
      bandNames.push_back("RAPID_EYE_NIR");

      bandNames.push_back("LISS3_GREEN");
      bandNames.push_back("LISS3_RED");
      bandNames.push_back("LISS3_NEAR_INFRARED");
      bandNames.push_back("LISS3_SHORTWAVE_INFRARED");

      bandNames.push_back("DMC_NIR");
      bandNames.push_back("DMC_RED");
      bandNames.push_back("DMC_GREEN");

      bandNames.push_back("LANDSAT8-OLI1_COASTAL_AEROSOL");
      bandNames.push_back("LANDSAT8-OLI2_BLUE");
      bandNames.push_back("LANDSAT8-OLI3_GREEN");
      bandNames.push_back("LANDSAT8-OLI4_RED");
      bandNames.push_back("LANDSAT8-OLI5_NIR");
      bandNames.push_back("LANDSAT8-OLI6_SWIR1");
      bandNames.push_back("LANDSAT8-OLI7_SWIR2");
      bandNames.push_back("LANDSAT8-OLI8_PAN");
      bandNames.push_back("LANDSAT8-OLI9_CIRRUS");

      bandNames.push_back("LANDSAT8-TIRS1");
      bandNames.push_back("LANDSAT8-TIRS2");

      bandNames.push_back("CBERS4_PAN_PAN");
      bandNames.push_back("CBERS4_PAN_GREEN");
      bandNames.push_back("CBERS4_PAN_RED");
      bandNames.push_back("CBERS4_PAN_NIR");

      bandNames.push_back("CBERS4_MUX_BLUE");
      bandNames.push_back("CBERS4_MUX_GREEN");
      bandNames.push_back("CBERS4_MUX_RED");
      bandNames.push_back("CBERS4_MUX_NIR");

      bandNames.push_back("CBERS4_IRS_PAN");
      bandNames.push_back("CBERS4_IRS_SWIR_B10");
      bandNames.push_back("CBERS4_IRS_SWIR_B11");
      bandNames.push_back("CBERS4_IRS_TH");

      bandNames.push_back("CBERS4_WFI_BLUE");
      bandNames.push_back("CBERS4_WFI_GREEN");
      bandNames.push_back("CBERS4_WFI_RED");
      bandNames.push_back("CBERS4_WFI_NIR");

      return bandNames;
    }

    std::map<std::string, SpectralSensorParams > getSensorParams()
    {
      std::string sens_file = GetSensorFilename();

      static std::map<std::string, SpectralSensorParams > SensorParams;
      SensorParams.clear();

      if (!boost::filesystem::exists(sens_file)) //SpectralSensor.json doesn't exists, then uses the original sensors
      {
        std::vector<std::string> bn = GetBandNames();
        for (size_t i = 0; i < bn.size(); i++)
          SensorParams[bn[i]] = GetSpectralBandInfo(bn[i]);
      }
      else
      {
        boost::property_tree::ptree pt;
        boost::property_tree::json_parser::read_json(sens_file, pt);

        for(boost::property_tree::ptree::value_type &v: pt.get_child("Spectral Sensors"))
        {
          std::string name = v.second.get<std::string>("name");
          std::string band = v.second.get<std::string>("band");
          std::string lower = v.second.get<std::string>("lower");
          std::string upper = v.second.get<std::string>("upper");
          std::string min = v.second.get<std::string>("min");
          std::string max = v.second.get<std::string>("max");
          SensorParams[name] = SpectralSensorParams(std::atoi(band.c_str()), std::atof(lower.c_str()), std::atof(upper.c_str()), std::atof(min.c_str()), std::atof(max.c_str()));
        }
      }

      return SensorParams;
    }

    void SaveSensorParams(std::map<std::string, SpectralSensorParams > &params)
    {
      std::string filename = GetSensorFilename();

      boost::property_tree::ptree pt;
      boost::property_tree::ptree sensors;

      for (std::map<std::string, te::rp::SpectralSensorParams > ::iterator it = params.begin(); it != params.end(); ++it)
      {
        boost::property_tree::ptree sensor;

        sensor.put("name", it->first);
        sensor.put("band", it->second.m_band);
        sensor.put("lower", it->second.m_lower);
        sensor.put("upper", it->second.m_upper);
        sensor.put("min", it->second.m_min);
        sensor.put("max", it->second.m_max);

        sensors.push_back(std::make_pair("Spectral Sensor", sensor));
      }

      pt.add_child("Spectral Sensors", sensors);

      boost::property_tree::json_parser::write_json(filename, pt);

    }

    SpectralSensorParams GetSpectralBandInfo(std::string bandName)
    {
      static std::map<std::string, struct SpectralSensorParams > BandInfo;
      BandInfo.clear();

      BandInfo["CBERS2_CCD_1_BLUE"] = SpectralSensorParams(1, 0.45, 0.52, 0.0, 255.0);
      BandInfo["CBERS2_CCD_2_GREEN"] = SpectralSensorParams(2, 0.52, 0.59, 0.0, 255.0);
      BandInfo["CBERS2_CCD_3_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
      BandInfo["CBERS2_CCD_4_NIR"] = SpectralSensorParams(4, 0.77, 0.89, 0.0, 255.0);
      BandInfo["CBERS2_CCD_5_PAN"] = SpectralSensorParams(5, 0.51, 0.73, 0.0, 255.0);

      BandInfo["LANDSAT5_TM_1_BLUE"] = SpectralSensorParams(1, 0.45, 0.52, 0.0, 255.0);
      BandInfo["LANDSAT5_TM_2_GREEN"] = SpectralSensorParams(2, 0.52, 0.60, 0.0, 255.0);
      BandInfo["LANDSAT5_TM_3_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
      BandInfo["LANDSAT5_TM_4_NIR"] = SpectralSensorParams(4, 0.76, 0.90, 0.0, 255.0);
      BandInfo["LANDSAT5_TM_5_SWIR"] = SpectralSensorParams(5, 1.55, 1.75, 0.0, 255.0);
      BandInfo["LANDSAT5_TM_6_THERMAL"] = SpectralSensorParams(6, 10.40, 12.50, 0.0, 255.0);
      BandInfo["LANDSAT5_TM_7_MIR"] = SpectralSensorParams(7, 2.08, 2.35, 0.0, 255.0);

      BandInfo["LANDSAT7_ETM+_1_BLUE"] = SpectralSensorParams(1, 0.45, 0.515, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_2_GREEN"] = SpectralSensorParams(2, 0.525, 0.605, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_3_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_4_NIR"] = SpectralSensorParams(4, 0.75, 0.90, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_5_SWIR"] = SpectralSensorParams(5, 1.55, 1.75, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_6_THERMAL"] = SpectralSensorParams(6, 10.40, 12.5, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_7_MIR"] = SpectralSensorParams(7, 2.09, 2.35, 0.0, 255.0);
      BandInfo["LANDSAT7_ETM+_8_PAN"] = SpectralSensorParams(8, 0.52, 0.90, 0.0, 255.0);

      BandInfo["WV2_MUL_1_COASTAL"] = SpectralSensorParams(1, 0.4, 0.45, 0.0, 2048.0);
      BandInfo["WV2_MUL_2_BLUE"] = SpectralSensorParams(2, 0.45, 0.51, 0.0, 2048.0);
      BandInfo["WV2_MUL_3_GREEN"] = SpectralSensorParams(3, 0.51, 0.58, 0.0, 2048.0);
      BandInfo["WV2_MUL_4_YELLOW"] = SpectralSensorParams(4, 0.585, 0.625, 0.0, 2048.0);
      BandInfo["WV2_MUL_5_RED"] = SpectralSensorParams(5, 0.66, 0.69, 0.0, 2048.0);
      BandInfo["WV2_MUL_6_REDEDGE"] = SpectralSensorParams(6, 0.705, 0.745, 0.0, 2048.0);
      BandInfo["WV2_MUL_7_NIR1"] = SpectralSensorParams(7, 0.77, 0.895, 0.0, 2048.0);
      BandInfo["WV2_MUL_8_NIR2"] = SpectralSensorParams(8, 0.86, 0.104, 0.0, 2048.0);
      BandInfo["WV2_PAN"] = SpectralSensorParams(9, 0.45, 0.8, 0.0, 2048.0);

      BandInfo["AWIFS_RED"] = SpectralSensorParams(1, 0.62, 0.68, 0.0, 4096.0);
      BandInfo["AWIFS_GREEN"] = SpectralSensorParams(2, 0.52, 0.59, 0.0, 4096.0);
      BandInfo["AWIFS_NIR"] = SpectralSensorParams(3, 0.77, 0.86, 0.0, 4096.0);
      BandInfo["AWIFS_MIR"] = SpectralSensorParams(4, 1.55, 1.70, 0.0, 4096.0);

      BandInfo["RAPID_EYE_BLUE"] = SpectralSensorParams(1, 440.00, 510.00, 0.0, 4096.0);
      BandInfo["RAPID_EYE_GREEN"] = SpectralSensorParams(2, 520.00, 590.00, 0.0, 4096.0);
      BandInfo["RAPID_EYE_RED"] = SpectralSensorParams(3, 630.00, 685.00, 0.0, 4096.0);
      BandInfo["RAPID_EYE_REDEDGE"] = SpectralSensorParams(4, 690.00, 730.00, 0.0, 4096.0);
      BandInfo["RAPID_EYE_NIR"] = SpectralSensorParams(5, 760.00, 850.00, 0.0, 4096.0);

      BandInfo["LISS3_GREEN"] = SpectralSensorParams(1, 0.52, 0.59, 0.0, 128.0);
      BandInfo["LISS3_RED"] = SpectralSensorParams(2, 0.62, 0.68, 0.0, 128.0);
      BandInfo["LISS3_NEAR_INFRARED"] = SpectralSensorParams(3, 0.77, 0.86, 0.0, 128.0);
      BandInfo["LISS3_SHORTWAVE_INFRARED"] = SpectralSensorParams(4, 1.55, 1.70, 0.0, 128.0);

      BandInfo["DMC_NIR"] = SpectralSensorParams(1, 0.77, 0.90, 0.0, 1024.0);
      BandInfo["DMC_RED"] = SpectralSensorParams(2, 0.63, 0.69, 0.0, 1024.0);
      BandInfo["DMC_GREEN"] = SpectralSensorParams(3, 0.52, 0.60, 0.0, 1024.0);

      BandInfo["LANDSAT8-OLI1_COASTAL_AEROSOL"] = SpectralSensorParams(1, 0.43, 0.45, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI2_BLUE"] = SpectralSensorParams(2, 0.45, 0.51, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI3_GREEN"] = SpectralSensorParams(3, 0.53, 0.59, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI4_RED"] = SpectralSensorParams(4, 0.64, 0.67, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI5_NIR"] = SpectralSensorParams(5, 0.85, 0.88, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI6_SWIR1"] = SpectralSensorParams(6, 1.57, 1.65, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI7_SWIR2"] = SpectralSensorParams(7, 2.11, 2.29, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI8_PAN"] = SpectralSensorParams(8, 0.50, 0.68, 0.0, 65536.0);
      BandInfo["LANDSAT8-OLI9_CIRRUS"] = SpectralSensorParams(9, 1.36, 1.38, 0.0, 65536.0);

      BandInfo["LANDSAT8-TIRS1"] = SpectralSensorParams(10, 10.60, 11.19, 0.0, 65536.0);
      BandInfo["LANDSAT8-TIRS2"] = SpectralSensorParams(11, 11.50, 12.51, 0.0, 65536.0);

      BandInfo["CBERS4_PAN_PAN"] = SpectralSensorParams(1, 0.51, 0.85, 0.0, 255.0);
      BandInfo["CBERS4_PAN_GREEN"] = SpectralSensorParams(2, 0.52, 0.59, 0.0, 255.0);
      BandInfo["CBERS4_PAN_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
      BandInfo["CBERS4_PAN_NIR"] = SpectralSensorParams(4, 0.77, 0.89, 0.0, 255.0);

      BandInfo["CBERS4_MUX_BLUE"] = SpectralSensorParams(5, 0.45, 0.52, 0.0, 255.0);
      BandInfo["CBERS4_MUX_GREEN"] = SpectralSensorParams(6, 0.52, 0.59, 0.0, 255.0);
      BandInfo["CBERS4_MUX_RED"] = SpectralSensorParams(7, 0.63, 0.69, 0.0, 255.0);
      BandInfo["CBERS4_MUX_NIR"] = SpectralSensorParams(8, 0.77, 0.89, 0.0, 255.0);

      BandInfo["CBERS4_IRS_PAN"] = SpectralSensorParams(9, 0.50, 0.90, 0.0, 255.0);
      BandInfo["CBERS4_IRS_SWIR_B10"] = SpectralSensorParams(10, 1.55, 1.75, 0.0, 255.0);
      BandInfo["CBERS4_IRS_SWIR_B11"] = SpectralSensorParams(11, 2.08, 2.35, 0.0, 255.0);
      BandInfo["CBERS4_IRS_TH"] = SpectralSensorParams(12, 10.40, 12.50, 0.0, 255.0);

      BandInfo["CBERS4_WFI_BLUE"] = SpectralSensorParams(13, 0.45, 0.52, 0.0, 1024.0);
      BandInfo["CBERS4_WFI_GREEN"] = SpectralSensorParams(14, 0.52, 0.59, 0.0, 1024.0);
      BandInfo["CBERS4_WFI_RED"] = SpectralSensorParams(15, 0.63, 0.69, 0.0, 1024.0);
      BandInfo["CBERS4_WFI_NIR"] = SpectralSensorParams(16, 0.77, 0.89, 0.0, 1024.0);


      // needs more Band Info from other sensors/bands
      if (BandInfo.find(bandName) == BandInfo.end())
        return SpectralSensorParams(0, 0.0, 1.0, 0.0, 255.0);

      return BandInfo[bandName];
    }

    std::pair<double, double> GetDigitalNumberBandInfo(std::string bandName)
    {
      std::map<std::string, SpectralSensorParams > params = getSensorParams();
      std::map<std::string, SpectralSensorParams > ::iterator it = params.find(bandName);
      if (it == params.end())
        return std::pair<double, double>(0.0, 255.0);

      return std::pair<double, double>(it->second.m_min, it->second.m_max);
    }

    double GetDigitalNumberBandMax(std::string bandName)
    {
      return GetDigitalNumberBandInfo(bandName).second;
    }

    double GetDigitalNumberBandMin(std::string bandName)
    {
      return GetDigitalNumberBandInfo(bandName).first;
    }

    bool NormalizeRaster(te::rst::Raster& inraster, double nmin, double nmax)
    {
      if (nmin > nmax)
        return false;

    // computing old maximuns and minimums for each band
      std::vector<double> omins;
      std::vector<double> omaxs;
      std::vector<unsigned int> bands;
      std::vector<double> dummy;

      for (unsigned int b = 0; b < inraster.getNumberOfBands(); b++)
      {
        omins.push_back(inraster.getBand(b)->getMinValue(true).real());
        omaxs.push_back(inraster.getBand(b)->getMaxValue(true).real());
        dummy.push_back(inraster.getBand(b)->getProperty()->m_noDataValue);

        bands.push_back(b);
      }

    // calculating amplitudes to avoid multiple subtractions
      double value;
      const double namplitude = nmax - nmin;
      std::vector<double> oamplitude;
      for (unsigned int b = 0; b < inraster.getNumberOfBands(); b++)
        oamplitude.push_back(omaxs[b] - omins[b]);

    // iterating over raster to normalize pixel values
      te::rst::RasterIterator<double> it = te::rst::RasterIterator<double>::begin(&inraster, bands);
      te::rst::RasterIterator<double> itend = te::rst::RasterIterator<double>::end(&inraster, bands);

      while (it != itend)
      {
        for (unsigned int b = 0; b < inraster.getNumberOfBands(); b++)
        {
          value = (*it)[b];
          if (value != dummy[b])
            value = ((*it)[b] - omins[b]) * namplitude / oamplitude[b] + nmin;

          inraster.setValue(it.getColumn(), it.getRow(), value, b);
        }

        ++it;
      }

      return true;
    }

    te::rst::Raster* NormalizeRaster(const te::rst::Raster* inputRaster)
    {
      double nmin = 0.0;
      double nmax = 255.0;

      // computing old maximuns and minimums for each band
      std::vector<double> omins;
      std::vector<double> omaxs;
      std::vector<unsigned int> bands;
      std::vector<double> dummy;

      for (unsigned int b = 0; b < inputRaster->getNumberOfBands(); b++)
      {
        omins.push_back(inputRaster->getBand(b)->getMinValue(true).real());
        omaxs.push_back(inputRaster->getBand(b)->getMaxValue(true).real());
        dummy.push_back(inputRaster->getBand(b)->getProperty()->m_noDataValue);

        bands.push_back(b);
      }

      // calculating amplitudes to avoid multiple subtractions
      double value;
      const double namplitude = nmax - nmin;
      std::vector<double> oamplitude;
      for (unsigned int b = 0; b < inputRaster->getNumberOfBands(); b++)
        oamplitude.push_back(omaxs[b] - omins[b]);

      // iterating over raster to normalize pixel values
      te::rst::RasterIterator<double> it = te::rst::RasterIterator<double>::begin((te::rst::Raster*)inputRaster, bands);
      te::rst::RasterIterator<double> itend = te::rst::RasterIterator<double>::end((te::rst::Raster*)inputRaster, bands);

      std::vector<te::rst::BandProperty*> bandsProperties;

      for (unsigned i = 0; i < inputRaster->getNumberOfBands(); i++)
      {
        te::rst::BandProperty* bprop = new te::rst::BandProperty(*(inputRaster->getBand(i)->getProperty()));
        bprop->m_type = te::dt::UCHAR_TYPE;
        bprop->m_noDataValue = inputRaster->getBand(i)->getProperty()->m_noDataValue;
        bandsProperties.push_back(bprop);
      }

      std::map< std::string, std::string > rInfo;

      te::rst::Raster* outputRaster = te::rst::RasterFactory::make("MEM", new te::rst::Grid(*inputRaster->getGrid()), bandsProperties, rInfo);

      while (it != itend)
      {
        for (unsigned int b = 0; b < inputRaster->getNumberOfBands(); b++)
        {
          value = (*it)[b];
          if (value != dummy[b])
            value = ((*it)[b] - omins[b]) * namplitude / oamplitude[b] + nmin;

          outputRaster->setValue(it.getColumn(), it.getRow(), value, b);
        }

        ++it;
      }

      double rasterValue;

      for (unsigned r = 0; r < inputRaster->getNumberOfRows(); r++)
      {
        for (unsigned c = 0; c < inputRaster->getNumberOfColumns(); c++)
        {
          for (unsigned b = 0; b < inputRaster->getNumberOfBands(); b++)
          {
            inputRaster->getValue(c, r, rasterValue, b);

            if (rasterValue != inputRaster->getBand(b)->getProperty()->m_noDataValue)
               rasterValue = 255 * (rasterValue - omins[b]) / (omaxs[b] - omins[b]);

            outputRaster->setValue(c, r, rasterValue, b);
          }
        }
      }

      return outputRaster;
    }
  
    bool ConvertRGB2IHS( const te::rst::Raster& inputRGBRaster, 
      const unsigned int redBandIdx, const unsigned int greenBandIdx,
      const unsigned int blueBandIdx, const double rgbRangeMin, 
      const double rgbRangeMax, te::rst::Raster& outputIHSRaster ) 
    {
      if( ( inputRGBRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
        return false;
      if( redBandIdx > inputRGBRaster.getNumberOfBands() ) return false;
      if( greenBandIdx > inputRGBRaster.getNumberOfBands() ) return false;
      if( blueBandIdx > inputRGBRaster.getNumberOfBands() ) return false;
      if( ( outputIHSRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
        return false;      
      if( outputIHSRaster.getNumberOfBands() < 3 ) return false;
      if( inputRGBRaster.getNumberOfColumns() != outputIHSRaster.getNumberOfColumns() )
        return false;
      if( inputRGBRaster.getNumberOfRows() != outputIHSRaster.getNumberOfRows() )
        return false;      
      
      const te::rst::Band& redBand = *inputRGBRaster.getBand( redBandIdx );
      const te::rst::Band& greenBand = *inputRGBRaster.getBand( greenBandIdx );
      const te::rst::Band& blueBand = *inputRGBRaster.getBand( blueBandIdx ); 
      
      te::rst::Band& iBand = *outputIHSRaster.getBand( 0 );
      te::rst::Band& hBand = *outputIHSRaster.getBand( 1 );
      te::rst::Band& sBand = *outputIHSRaster.getBand( 2 );
      
      const unsigned int outNCols = outputIHSRaster.getNumberOfColumns();
      const unsigned int outNRows = outputIHSRaster.getNumberOfRows();
      const double redNoData = redBand.getProperty()->m_noDataValue;
      const double greenNoData = greenBand.getProperty()->m_noDataValue;
      const double blueNoData = inputRGBRaster.getBand( 
        blueBandIdx )->getProperty()->m_noDataValue;
      const double iNoData = iBand.getProperty()->m_noDataValue;
      const double hNoData = hBand.getProperty()->m_noDataValue;
      const double sNoData = sBand.getProperty()->m_noDataValue;
      
      unsigned int outRow = 0;
      unsigned int outCol = 0;
      double red = 0;
      double green = 0;
      double blue = 0;
      double teta = 0;
      double redNorm = 0, greenNorm = 0, blueNorm = 0;
      double rMinusG = 0, rMinusB = 0;
      double rgbSum = 0;
      double cosValue = 0;
      const double twoPi = 2.0 * (static_cast<double>(M_PI));
      const double pi2 = (static_cast<double>(M_PI)) / 2.0;
      const double rgbNormFac = ( rgbRangeMax == rgbRangeMin ) ? 0.0 :
        ( 1.0 / ( rgbRangeMax - rgbRangeMin ) );
      double intensity = 0;
      double hue = 0;
      double saturation = 0;
      
      for( outRow = 0 ; outRow < outNRows ; ++outRow )
      {
        for( outCol = 0 ; outCol < outNCols ; ++outCol )
        {
          redBand.getValue( outCol, outRow, red );
          greenBand.getValue( outCol, outRow, green );
          blueBand.getValue( outCol, outRow, blue );
          
          if( ( red == redNoData ) || ( green == greenNoData ) ||
            ( blue == blueNoData ) )
          {
            iBand.setValue( outCol, outRow, iNoData );
            hBand.setValue( outCol, outRow, hNoData );
            sBand.setValue( outCol, outRow, sNoData );
          }            
          else
          {
            if( ( red == green ) && ( green == blue ) ) 
            { // Gray scale case
              // From Wikipedia:
              // h = 0 is used for grays though the hue has no geometric 
              // meaning there, where the saturation s = 0. Similarly, 
              // the choice of 0 as the value for s when l is equal to 0 or 1 
              // is arbitrary.        
              
              hue = 0.0;
              saturation = 0.0;
              intensity = ( red * rgbNormFac ); // or green or blue since they all are the same.
            }
            else
            { // Color case
              redNorm = ( red - rgbRangeMin ) * rgbNormFac;
              greenNorm = ( green - rgbRangeMin ) * rgbNormFac;
              blueNorm = ( blue - rgbRangeMin ) * rgbNormFac;
              
              rMinusG = redNorm - greenNorm;
              rMinusB = redNorm - blueNorm;
              
              cosValue = std::sqrt( ( rMinusG * rMinusG ) + ( rMinusB * 
                ( greenNorm - blueNorm ) ) );
                
              if( cosValue == 0.0 )
              {
                teta = pi2;
              }
              else
              {
                cosValue =  ( 0.5 * ( rMinusG + rMinusB )  ) /
                  cosValue;
                teta = std::acos( cosValue );  
              }
                
              assert( ( cosValue >= (-1.0) ) && ( cosValue <= (1.0) ) );
                
              if( blueNorm > greenNorm )
              {
                hue = ( twoPi - teta );
              }
              else
              {
                hue = teta;
              }
                
              rgbSum = ( redNorm + greenNorm + blueNorm );
              
              saturation = ( 1.0 - ( 3 * std::min( std::min( redNorm, greenNorm ), blueNorm ) /
                rgbSum ) );
                
              intensity = ( rgbSum / 3.0 );            
            }
            
            iBand.setValue( outCol, outRow, intensity );
            hBand.setValue( outCol, outRow, hue );
            sBand.setValue( outCol, outRow, saturation );            
          }
        }
      }      
      
      return true;
    }

    bool ConvertRGB2IHS(const te::rst::Raster& inputRedRaster, const unsigned int redBandIdx,
      const te::rst::Raster& inputGreenRaster, const unsigned int greenBandIdx,
      const te::rst::Raster& inputBlueRaster, const unsigned int blueBandIdx,
      const double rgbRangeMin, const double rgbRangeMax, te::rst::Raster& outputIHSRaster)
    {
      if ((inputRedRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputGreenRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputBlueRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (redBandIdx > inputRedRaster.getNumberOfBands()) return false;
      if (greenBandIdx > inputGreenRaster.getNumberOfBands()) return false;
      if (blueBandIdx > inputBlueRaster.getNumberOfBands()) return false;
      if ((outputIHSRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputIHSRaster.getNumberOfBands() < 3) return false;
      if (inputRedRaster.getNumberOfColumns() != inputGreenRaster.getNumberOfColumns() ||
          inputRedRaster.getNumberOfColumns() != inputBlueRaster.getNumberOfColumns())
        return false;
      if (inputRedRaster.getNumberOfRows() != inputGreenRaster.getNumberOfRows() ||
          inputRedRaster.getNumberOfRows() != inputBlueRaster.getNumberOfRows())
        return false;
      if (inputRedRaster.getNumberOfColumns() != outputIHSRaster.getNumberOfColumns())
        return false;
      if (inputRedRaster.getNumberOfRows() != outputIHSRaster.getNumberOfRows())
        return false;

      const te::rst::Band& redBand = *inputRedRaster.getBand(redBandIdx);
      const te::rst::Band& greenBand = *inputGreenRaster.getBand(greenBandIdx);
      const te::rst::Band& blueBand = *inputBlueRaster.getBand(blueBandIdx);

      te::rst::Band& iBand = *outputIHSRaster.getBand(0);
      te::rst::Band& hBand = *outputIHSRaster.getBand(1);
      te::rst::Band& sBand = *outputIHSRaster.getBand(2);

      const unsigned int outNCols = outputIHSRaster.getNumberOfColumns();
      const unsigned int outNRows = outputIHSRaster.getNumberOfRows();
      const double redNoData = redBand.getProperty()->m_noDataValue;
      const double greenNoData = greenBand.getProperty()->m_noDataValue;
      const double blueNoData = blueBand.getProperty()->m_noDataValue;
      const double iNoData = iBand.getProperty()->m_noDataValue;
      const double hNoData = hBand.getProperty()->m_noDataValue;
      const double sNoData = sBand.getProperty()->m_noDataValue;

      unsigned int outRow = 0;
      unsigned int outCol = 0;
      double red = 0;
      double green = 0;
      double blue = 0;
      double teta = 0;
      double redNorm = 0, greenNorm = 0, blueNorm = 0;
      double rMinusG = 0, rMinusB = 0;
      double rgbSum = 0;
      double cosValue = 0;
      const double twoPi = 2.0 * (static_cast<double>(M_PI));
      const double pi2 = (static_cast<double>(M_PI)) / 2.0;
      const double rgbNormFac = (rgbRangeMax == rgbRangeMin) ? 0.0 :
        (1.0 / (rgbRangeMax - rgbRangeMin));
      double intensity = 0;
      double hue = 0;
      double saturation = 0;

      for (outRow = 0; outRow < outNRows; ++outRow)
      {
        for (outCol = 0; outCol < outNCols; ++outCol)
        {
          redBand.getValue(outCol, outRow, red);
          greenBand.getValue(outCol, outRow, green);
          blueBand.getValue(outCol, outRow, blue);

          if ((red == redNoData) || (green == greenNoData) ||
            (blue == blueNoData))
          {
            iBand.setValue(outCol, outRow, iNoData);
            hBand.setValue(outCol, outRow, hNoData);
            sBand.setValue(outCol, outRow, sNoData);
          }
          else
          {
            if ((red == green) && (green == blue))
            { // Gray scale case
              // From Wikipedia:
              // h = 0 is used for grays though the hue has no geometric 
              // meaning there, where the saturation s = 0. Similarly, 
              // the choice of 0 as the value for s when l is equal to 0 or 1 
              // is arbitrary.        

              hue = 0.0;
              saturation = 0.0;
              intensity = (red * rgbNormFac); // or green or blue since they all are the same.
            }
            else
            { // Color case
              redNorm = (red - rgbRangeMin) * rgbNormFac;
              greenNorm = (green - rgbRangeMin) * rgbNormFac;
              blueNorm = (blue - rgbRangeMin) * rgbNormFac;

              rMinusG = redNorm - greenNorm;
              rMinusB = redNorm - blueNorm;

              cosValue = std::sqrt((rMinusG * rMinusG) + (rMinusB *
                (greenNorm - blueNorm)));

              if (cosValue == 0.0)
              {
                teta = pi2;
              }
              else
              {
                cosValue = (0.5 * (rMinusG + rMinusB)) /
                  cosValue;
                teta = std::acos(cosValue);
              }

              assert((cosValue >= (-1.0)) && (cosValue <= (1.0)));

              if (blueNorm > greenNorm)
              {
                hue = (twoPi - teta);
              }
              else
              {
                hue = teta;
              }

              rgbSum = (redNorm + greenNorm + blueNorm);

              saturation = (1.0 - (3 * std::min(std::min(redNorm, greenNorm), blueNorm) /
                rgbSum));

              intensity = (rgbSum / 3.0);
            }

            iBand.setValue(outCol, outRow, intensity);
            hBand.setValue(outCol, outRow, hue);
            sBand.setValue(outCol, outRow, saturation);
          }
        }
      }

      return true;
    }
    
    bool ConvertIHS2RGB( const te::rst::Raster& inputIHSRaster, 
      const unsigned int intensityBandIdx, const unsigned int hueBandIdx,
      const unsigned int saturationBandIdx, const double rgbRangeMin, 
      const double rgbRangeMax, te::rst::Raster& outputRGBRaster )
    {
      if( ( inputIHSRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
        return false;
      if( intensityBandIdx > inputIHSRaster.getNumberOfBands() ) return false;
      if( hueBandIdx > inputIHSRaster.getNumberOfBands() ) return false;
      if( saturationBandIdx > inputIHSRaster.getNumberOfBands() ) return false;
      if( ( outputRGBRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
        return false;      
      if( outputRGBRaster.getNumberOfBands() < 3 ) return false;
      if( inputIHSRaster.getNumberOfColumns() != outputRGBRaster.getNumberOfColumns() )
        return false;
      if( inputIHSRaster.getNumberOfRows() != outputRGBRaster.getNumberOfRows() )
        return false;      
      
      const te::rst::Band& intensityBand = *inputIHSRaster.getBand( intensityBandIdx );
      const te::rst::Band& hueBand = *inputIHSRaster.getBand( hueBandIdx );
      const te::rst::Band& saturationBand = *inputIHSRaster.getBand( saturationBandIdx );      
      te::rst::Band& redBand = *outputRGBRaster.getBand( 0 );
      te::rst::Band& greenBand = *outputRGBRaster.getBand( 1 );
      te::rst::Band& blueBand = *outputRGBRaster.getBand( 2 );      
      
      const unsigned int nCols = outputRGBRaster.getNumberOfColumns();
      const unsigned int nRows = outputRGBRaster.getNumberOfRows();
      
      const double intensityNoData = intensityBand.getProperty()->m_noDataValue;
      const double hueNoData = hueBand.getProperty()->m_noDataValue;
      const double saturationNoData = saturationBand.getProperty()->m_noDataValue;
        
      const double rNoData = redBand.getProperty()->m_noDataValue;
      const double gNoData = greenBand.getProperty()->m_noDataValue;
      const double bNoData = blueBand.getProperty()->m_noDataValue;
        
      const double rgbNormFac = ( rgbRangeMax == rgbRangeMin ) ? 0.0 :
        ( rgbRangeMax - rgbRangeMin ); 
      const double pi3 = M_PI / 3.0; // 60
      const double twoPi3 = 2.0 * M_PI / 3.0; // 120
      const double fourPi3 = 4.0 * M_PI / 3.0; // 240        
      unsigned int row = 0;
      unsigned int col = 0;  
      double hue = 0;
      double lig = 0;
      double sat = 0;
      double red = 0;
      double green = 0;
      double blue = 0;  

      
      for( row = 0 ; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {     
          intensityBand.getValue( col, row, lig );
          hueBand.getValue( col, row, hue );
          saturationBand.getValue( col, row, sat );
          
          if( ( lig == intensityNoData ) || ( hue == hueNoData ) || 
            ( sat == saturationNoData ) )
          {
            redBand.setValue( col, row, rNoData );
            greenBand.setValue( col, row, gNoData );
            blueBand.setValue( col, row, bNoData );            
          }
          else
          {
            if( ( hue == 0.0 ) && ( sat == 0.0 ) )
            { // Gray scale case
              red = green = blue = ( lig * rgbNormFac );
            }
            else
            { // color case
              /* Hue inside RG sector */
              if( hue < twoPi3 )
              {
                blue = lig * ( 1.0 - sat );
                red = lig * ( 1.0 + ( sat * std::cos( hue ) / 
                  std::cos( pi3 - hue ) ) );
                green = ( 3.0 * lig ) - ( red + blue );
              }
              else if( hue < fourPi3 )
              { /* Hue inside GB sector */
              
                hue -= twoPi3;
                
                red = lig * ( 1.0 - sat );
                green = lig * ( 1.0 + ( sat * std::cos( hue ) / 
                  std::cos( pi3 - hue ) ) );
                blue = ( 3.0 * lig ) - ( red + green );
              }
              else
              { /* Hue inside BR sector */
              
                hue -= fourPi3;
                
                green = lig * ( 1.0 - sat );
                blue = lig * ( 1.0 + ( sat * std::cos( hue ) / 
                  std::cos( pi3 - hue ) ) );
                red = ( 3.0 * lig ) - ( green + blue );
              }
              
              red = ( red * rgbNormFac ) + rgbRangeMin;
              green = ( green * rgbNormFac ) + rgbRangeMin;
              blue = ( blue * rgbNormFac ) + rgbRangeMin;
            }
            
            red = MIN( red, rgbRangeMax );
            green = MIN( green, rgbRangeMax );
            blue = MIN( blue, rgbRangeMax );
            
            red = MAX( red, rgbRangeMin );
            green = MAX( green, rgbRangeMin );
            blue = MAX( blue, rgbRangeMin );           
            
            redBand.setValue( col, row, red );
            greenBand.setValue( col, row, green );
            blueBand.setValue( col, row, blue );            
          }
        }
      }        
      
      return true;
    }

    bool ConvertIHS2RGB(const te::rst::Raster& inputIRaster, const unsigned int intensityBandIdx,
      const te::rst::Raster& inputHRaster, const unsigned int hueBandIdx,
      const te::rst::Raster& inputSRaster, const unsigned int saturationBandIdx,
      const double rgbRangeMin, const double rgbRangeMax, te::rst::Raster& outputRGBRaster)
    {
      if ((inputIRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputHRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputSRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (intensityBandIdx > inputIRaster.getNumberOfBands()) return false;
      if (hueBandIdx > inputHRaster.getNumberOfBands()) return false;
      if (saturationBandIdx > inputSRaster.getNumberOfBands()) return false;
      if (inputIRaster.getNumberOfColumns() != inputHRaster.getNumberOfColumns() ||
        inputIRaster.getNumberOfColumns() != inputSRaster.getNumberOfColumns())
        return false;
      if (inputIRaster.getNumberOfRows() != inputHRaster.getNumberOfRows() ||
        inputIRaster.getNumberOfRows() != inputSRaster.getNumberOfRows())
        return false;
      if ((outputRGBRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputRGBRaster.getNumberOfBands() < 3) return false;
      if (inputIRaster.getNumberOfColumns() != outputRGBRaster.getNumberOfColumns())
        return false;
      if (inputIRaster.getNumberOfRows() != outputRGBRaster.getNumberOfRows())
        return false;

      const te::rst::Band& intensityBand = *inputIRaster.getBand(intensityBandIdx);
      const te::rst::Band& hueBand = *inputHRaster.getBand(hueBandIdx);
      const te::rst::Band& saturationBand = *inputSRaster.getBand(saturationBandIdx);
      te::rst::Band& redBand = *outputRGBRaster.getBand(0);
      te::rst::Band& greenBand = *outputRGBRaster.getBand(1);
      te::rst::Band& blueBand = *outputRGBRaster.getBand(2);

      const unsigned int nCols = outputRGBRaster.getNumberOfColumns();
      const unsigned int nRows = outputRGBRaster.getNumberOfRows();

      const double intensityNoData = intensityBand.getProperty()->m_noDataValue;
      const double hueNoData = hueBand.getProperty()->m_noDataValue;
      const double saturationNoData = saturationBand.getProperty()->m_noDataValue;

      const double rNoData = redBand.getProperty()->m_noDataValue;
      const double gNoData = greenBand.getProperty()->m_noDataValue;
      const double bNoData = blueBand.getProperty()->m_noDataValue;

      const double rgbNormFac = (rgbRangeMax == rgbRangeMin) ? 0.0 :
        (rgbRangeMax - rgbRangeMin);
      const double pi3 = M_PI / 3.0; // 60
      const double twoPi3 = 2.0 * M_PI / 3.0; // 120
      const double fourPi3 = 4.0 * M_PI / 3.0; // 240        
      unsigned int row = 0;
      unsigned int col = 0;
      double hue = 0;
      double lig = 0;
      double sat = 0;
      double red = 0;
      double green = 0;
      double blue = 0;


      for (row = 0; row < nRows; ++row)
      {
        for (col = 0; col < nCols; ++col)
        {
          intensityBand.getValue(col, row, lig);
          hueBand.getValue(col, row, hue);
          saturationBand.getValue(col, row, sat);

          if ((lig == intensityNoData) || (hue == hueNoData) ||
            (sat == saturationNoData))
          {
            redBand.setValue(col, row, rNoData);
            greenBand.setValue(col, row, gNoData);
            blueBand.setValue(col, row, bNoData);
          }
          else
          {
            if ((hue == 0.0) && (sat == 0.0))
            { // Gray scale case
              red = green = blue = (lig * rgbNormFac);
            }
            else
            { // color case
              /* Hue inside RG sector */
              if (hue < twoPi3)
              {
                blue = lig * (1.0 - sat);
                red = lig * (1.0 + (sat * std::cos(hue) /
                  std::cos(pi3 - hue)));
                green = (3.0 * lig) - (red + blue);
              }
              else if (hue < fourPi3)
              { /* Hue inside GB sector */

                hue -= twoPi3;

                red = lig * (1.0 - sat);
                green = lig * (1.0 + (sat * std::cos(hue) /
                  std::cos(pi3 - hue)));
                blue = (3.0 * lig) - (red + green);
              }
              else
              { /* Hue inside BR sector */

                hue -= fourPi3;

                green = lig * (1.0 - sat);
                blue = lig * (1.0 + (sat * std::cos(hue) /
                  std::cos(pi3 - hue)));
                red = (3.0 * lig) - (green + blue);
              }

              red = (red * rgbNormFac) + rgbRangeMin;
              green = (green * rgbNormFac) + rgbRangeMin;
              blue = (blue * rgbNormFac) + rgbRangeMin;
            }

            red = MIN(red, rgbRangeMax);
            green = MIN(green, rgbRangeMax);
            blue = MIN(blue, rgbRangeMax);

            red = MAX(red, rgbRangeMin);
            green = MAX(green, rgbRangeMin);
            blue = MAX(blue, rgbRangeMin);

            redBand.setValue(col, row, red);
            greenBand.setValue(col, row, green);
            blueBand.setValue(col, row, blue);
          }
        }
      }

      return true;
    }

    bool ConvertRGB2HLS(const te::rst::Raster& inputRGBRaster,
      const unsigned int redBandIdx, const unsigned int greenBandIdx,
      const unsigned int blueBandIdx, const double /*rgbRangeMin*/,
      const double rgbRangeMax, te::rst::Raster& outputHLSRaster)
    {
      if ((inputRGBRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (redBandIdx > inputRGBRaster.getNumberOfBands()) return false;
      if (greenBandIdx > inputRGBRaster.getNumberOfBands()) return false;
      if (blueBandIdx > inputRGBRaster.getNumberOfBands()) return false;
      if ((outputHLSRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputHLSRaster.getNumberOfBands() < 3) return false;
      if (inputRGBRaster.getNumberOfColumns() != outputHLSRaster.getNumberOfColumns())
        return false;
      if (inputRGBRaster.getNumberOfRows() != outputHLSRaster.getNumberOfRows())
        return false;

      const te::rst::Band& redBand = *inputRGBRaster.getBand(redBandIdx);
      const te::rst::Band& greenBand = *inputRGBRaster.getBand(greenBandIdx);
      const te::rst::Band& blueBand = *inputRGBRaster.getBand(blueBandIdx);

      te::rst::Band& hBand = *outputHLSRaster.getBand(0);
      te::rst::Band& lBand = *outputHLSRaster.getBand(1);
      te::rst::Band& sBand = *outputHLSRaster.getBand(2);

      const unsigned int outNCols = outputHLSRaster.getNumberOfColumns();
      const unsigned int outNRows = outputHLSRaster.getNumberOfRows();

      unsigned int outRow = 0;
      unsigned int outCol = 0;
      double red = 0;
      double green = 0;
      double blue = 0;
      double light = 0;
      double hue = 0;
      double saturation = 0;

      for (outRow = 0; outRow < outNRows; ++outRow)
      {
        for (outCol = 0; outCol < outNCols; ++outCol)
        {
          redBand.getValue(outCol, outRow, red);
          greenBand.getValue(outCol, outRow, green);
          blueBand.getValue(outCol, outRow, blue);

          red /= rgbRangeMax;
          green /= rgbRangeMax;
          blue /= rgbRangeMax;

          double vmax = MAX(red, green);
          vmax = MAX(vmax, blue);

          double vmin = MIN(red, green);
          vmin = MIN(vmin, blue);

          hue = (vmax + vmin) / 2.;
          saturation = (vmax + vmin) / 2.;
          light = (vmax + vmin) / 2.;

          if (vmax == vmin)
          {// achromatic
            hue = 0.;
            saturation = 0.;
          }
          else
          {
            double diff = vmax - vmin;

            saturation = light > 0.5 ? diff / (2. - vmax - vmin) : diff / (vmax + vmin);

            if (vmax == red)
            {
              hue = (green - blue) / diff + (green < blue ? 6. : 0.);
            }
            else if (vmax == green)
            {
              hue = (blue - red) / diff + 2.;
            }
            else if (vmax == blue)
            {
              hue = (red - green) / diff + 4.;
            }

            hue *= 60.;
          }

          hBand.setValue(outCol, outRow, hue);
          lBand.setValue(outCol, outRow, light);
          sBand.setValue(outCol, outRow, saturation);
        }
      }

      return true;
    }

    bool ConvertRGB2HLS(const te::rst::Raster& inputRedRaster, const unsigned int redBandIdx,
      const te::rst::Raster& inputGreenRaster, const unsigned int greenBandIdx,
      const te::rst::Raster& inputBlueRaster, const unsigned int blueBandIdx,
      const double /*rgbRangeMin*/, const double rgbRangeMax, te::rst::Raster& outputHLSRaster)
    {
      if ((inputRedRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputGreenRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputBlueRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (redBandIdx > inputRedRaster.getNumberOfBands()) return false;
      if (greenBandIdx > inputGreenRaster.getNumberOfBands()) return false;
      if (blueBandIdx > inputBlueRaster.getNumberOfBands()) return false;
      if ((outputHLSRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputHLSRaster.getNumberOfBands() < 3) return false;
      if (inputRedRaster.getNumberOfRows() != inputGreenRaster.getNumberOfRows() ||
        inputRedRaster.getNumberOfRows() != inputBlueRaster.getNumberOfRows())
        return false;
      if (inputRedRaster.getNumberOfColumns() != outputHLSRaster.getNumberOfColumns())
        return false;
      if (inputRedRaster.getNumberOfRows() != outputHLSRaster.getNumberOfRows())
        return false;

      const te::rst::Band& redBand = *inputRedRaster.getBand(redBandIdx);
      const te::rst::Band& greenBand = *inputGreenRaster.getBand(greenBandIdx);
      const te::rst::Band& blueBand = *inputBlueRaster.getBand(blueBandIdx);

      te::rst::Band& hBand = *outputHLSRaster.getBand(0);
      te::rst::Band& lBand = *outputHLSRaster.getBand(1);
      te::rst::Band& sBand = *outputHLSRaster.getBand(2);

      const unsigned int outNCols = outputHLSRaster.getNumberOfColumns();
      const unsigned int outNRows = outputHLSRaster.getNumberOfRows();

      unsigned int outRow = 0;
      unsigned int outCol = 0;
      double red = 0;
      double green = 0;
      double blue = 0;
      double light = 0;
      double hue = 0;
      double saturation = 0;

      for (outRow = 0; outRow < outNRows; ++outRow)
      {
        for (outCol = 0; outCol < outNCols; ++outCol)
        {
          redBand.getValue(outCol, outRow, red);
          greenBand.getValue(outCol, outRow, green);
          blueBand.getValue(outCol, outRow, blue);

          red /= rgbRangeMax;
          green /= rgbRangeMax;
          blue /= rgbRangeMax;

          double vmax = MAX(red, green);
          vmax = MAX(vmax, blue);

          double vmin = MIN(red, green);
          vmin = MIN(vmin, blue);

          hue = (vmax + vmin) / 2.;
          saturation = (vmax + vmin) / 2.;
          light = (vmax + vmin) / 2.;

          if (vmax == vmin)
          {// achromatic
            hue = 0.;
            saturation = 0.;
          }
          else
          {
            double diff = vmax - vmin;

            saturation = light > 0.5 ? diff / (2. - vmax - vmin) : diff / (vmax + vmin);

            if (vmax == red)
            {
              hue = (green - blue) / diff + (green < blue ? 6. : 0.);
            }
            else if (vmax == green)
            {
              hue = (blue - red) / diff + 2.;
            }
            else if (vmax == blue)
            {
              hue = (red - green) / diff + 4.;
            }

            hue *= 60.;
          }

          hBand.setValue(outCol, outRow, hue);
          lBand.setValue(outCol, outRow, light);
          sBand.setValue(outCol, outRow, saturation);
        }
      }

      return true;
    }

    bool ConvertHLS2RGB(const te::rst::Raster& inputHLSRaster,
      const unsigned int hueBandIdx, const unsigned int lightBandIdx,
      const unsigned int saturationBandIdx, const double /*rgbRangeMin*/,
      const double rgbRangeMax, te::rst::Raster& outputRGBRaster)
    {
      if ((inputHLSRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (hueBandIdx > inputHLSRaster.getNumberOfBands()) return false;
      if (lightBandIdx > inputHLSRaster.getNumberOfBands()) return false;
      if (saturationBandIdx > inputHLSRaster.getNumberOfBands()) return false;
      if ((outputRGBRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputRGBRaster.getNumberOfBands() < 3) return false;
      if (inputHLSRaster.getNumberOfColumns() != outputRGBRaster.getNumberOfColumns())
        return false;
      if (inputHLSRaster.getNumberOfRows() != outputRGBRaster.getNumberOfRows())
        return false;

      const te::rst::Band& hueBand = *inputHLSRaster.getBand(hueBandIdx);
      const te::rst::Band& lightBand = *inputHLSRaster.getBand(lightBandIdx);
      const te::rst::Band& saturationBand = *inputHLSRaster.getBand(saturationBandIdx);
      te::rst::Band& redBand = *outputRGBRaster.getBand(0);
      te::rst::Band& greenBand = *outputRGBRaster.getBand(1);
      te::rst::Band& blueBand = *outputRGBRaster.getBand(2);

      const unsigned int nCols = outputRGBRaster.getNumberOfColumns();
      const unsigned int nRows = outputRGBRaster.getNumberOfRows();

      const double hueNoData = hueBand.getProperty()->m_noDataValue;
      const double lightNoData = lightBand.getProperty()->m_noDataValue;
      const double saturationNoData = saturationBand.getProperty()->m_noDataValue;

      const double rNoData = redBand.getProperty()->m_noDataValue;
      const double gNoData = greenBand.getProperty()->m_noDataValue;
      const double bNoData = blueBand.getProperty()->m_noDataValue;
       
      unsigned int row = 0;
      unsigned int col = 0;
      double hue = 0;
      double light = 0;
      double saturation = 0;

      for (row = 0; row < nRows; ++row)
      {
        for (col = 0; col < nCols; ++col)
        {
          hueBand.getValue(col, row, hue);
          lightBand.getValue(col, row, light);
          saturationBand.getValue(col, row, saturation);

          if ((hue == hueNoData) || (light == lightNoData) ||
            (saturation == saturationNoData))
          {
            redBand.setValue(col, row, rNoData);
            greenBand.setValue(col, row, gNoData);
            blueBand.setValue(col, row, bNoData);
          }
          else
          {
            std::vector< double > rgb;
            hue = hue / 360.;

            if (saturation == 0)
            {// achromatic
              rgb.push_back(light * rgbRangeMax);
              rgb.push_back(light * rgbRangeMax);
              rgb.push_back(light * rgbRangeMax);
            }
            else
            {
              for (unsigned int i = 0; i < 3; ++i)
              {
                double q = light < 0.5 ? light * (1. + saturation) : light + saturation - light * saturation;
                double p = 2. * light - q;

                double t = 0;

                switch (i)
                {
                case 0://red
                {
                  t = hue + 1. / 3.;
                  break;
                }
                case 1://green
                {
                  t = hue;
                  break;
                }
                case 2://blue
                {
                  t = hue - 1. / 3.;
                  break;
                }
                }

                if (t < 0)
                {
                  t += 1.;
                }
                if (t > 1)
                {
                  t -= 1.;
                }

                if (t < (1. / 6.))
                {
                  p = (p + (q - p) * 6. * t);
                }
                else if (t < (1. / 2.))
                {
                  p = q;
                }
                else if (t < (2. / 3.))
                {
                  p = (p + (q - p) * (2. / 3. - t) * 6.);
                }

                rgb.push_back(p * rgbRangeMax);
              }
            }

            redBand.setValue(col, row, rgb[0]);
            greenBand.setValue(col, row, rgb[1]);
            blueBand.setValue(col, row, rgb[2]);
          }
        }
      }

      return true;
    }

    bool ConvertHLS2RGB(const te::rst::Raster& inputHRaster, const unsigned int hueBandIdx,
      const te::rst::Raster& inputLRaster, const unsigned int lightBandIdx,
      const te::rst::Raster& inputSRaster, const unsigned int saturationBandIdx,
      const double /*rgbRangeMin*/, const double rgbRangeMax, te::rst::Raster& outputRGBRaster)
    {
      if ((inputHRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputLRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputSRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (hueBandIdx > inputHRaster.getNumberOfBands()) return false;
      if (lightBandIdx > inputLRaster.getNumberOfBands()) return false;
      if (saturationBandIdx > inputSRaster.getNumberOfBands()) return false;
      if (inputHRaster.getNumberOfColumns() != inputLRaster.getNumberOfColumns() ||
          inputHRaster.getNumberOfColumns() != inputSRaster.getNumberOfColumns())
        return false;
      if (inputHRaster.getNumberOfRows() != inputLRaster.getNumberOfRows() ||
        inputHRaster.getNumberOfRows() != inputSRaster.getNumberOfRows())
        return false;
      if ((outputRGBRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputRGBRaster.getNumberOfBands() < 3) return false;
      if (inputHRaster.getNumberOfColumns() != outputRGBRaster.getNumberOfColumns())
        return false;
      if (inputHRaster.getNumberOfRows() != outputRGBRaster.getNumberOfRows())
        return false;

      const te::rst::Band& hueBand = *inputHRaster.getBand(hueBandIdx);
      const te::rst::Band& lightBand = *inputLRaster.getBand(lightBandIdx);
      const te::rst::Band& saturationBand = *inputSRaster.getBand(saturationBandIdx);
      te::rst::Band& redBand = *outputRGBRaster.getBand(0);
      te::rst::Band& greenBand = *outputRGBRaster.getBand(1);
      te::rst::Band& blueBand = *outputRGBRaster.getBand(2);

      const unsigned int nCols = outputRGBRaster.getNumberOfColumns();
      const unsigned int nRows = outputRGBRaster.getNumberOfRows();

      const double hueNoData = hueBand.getProperty()->m_noDataValue;
      const double lightNoData = lightBand.getProperty()->m_noDataValue;
      const double saturationNoData = saturationBand.getProperty()->m_noDataValue;

      const double rNoData = redBand.getProperty()->m_noDataValue;
      const double gNoData = greenBand.getProperty()->m_noDataValue;
      const double bNoData = blueBand.getProperty()->m_noDataValue;
   
      unsigned int row = 0;
      unsigned int col = 0;
      double hue = 0;
      double light = 0;
      double saturation = 0;

      for (row = 0; row < nRows; ++row)
      {
        for (col = 0; col < nCols; ++col)
        {
          hueBand.getValue(col, row, hue);
          lightBand.getValue(col, row, light);
          saturationBand.getValue(col, row, saturation);

          if ((hue == hueNoData) || (light == lightNoData) ||
            (saturation == saturationNoData))
          {
            redBand.setValue(col, row, rNoData);
            greenBand.setValue(col, row, gNoData);
            blueBand.setValue(col, row, bNoData);
          }
          else
          {
            std::vector< double > rgb;
            hue = hue / 360.;

            if (saturation == 0)
            {// achromatic
              rgb.push_back(light * rgbRangeMax);
              rgb.push_back(light * rgbRangeMax);
              rgb.push_back(light * rgbRangeMax);
            }
            else
            {
              for (unsigned int i = 0; i < 3; ++i)
              {
                double q = light < 0.5 ? light * (1. + saturation) : light + saturation - light * saturation;
                double p = 2. * light - q;

                double t = 0;

                switch (i)
                {
                case 0://red
                {
                  t = hue + 1. / 3.;
                  break;
                }
                case 1://green
                {
                  t = hue;
                  break;
                }
                case 2://blue
                {
                  t = hue - 1. / 3.;
                  break;
                }
                }

                if (t < 0)
                {
                  t += 1.;
                }
                if (t > 1)
                {
                  t -= 1.;
                }

                if (t < (1. / 6.))
                {
                  p = (p + (q - p) * 6. * t);
                }
                else if (t < (1. / 2.))
                {
                  p = q;
                }
                else if (t < (2. / 3.))
                {
                  p = (p + (q - p) * (2. / 3. - t) * 6.);
                }

                rgb.push_back(p * rgbRangeMax);
              }
            }

            redBand.setValue(col, row, rgb[0]);
            greenBand.setValue(col, row, rgb[1]);
            blueBand.setValue(col, row, rgb[2]);
          }
        }
      }

      return true;
    }
    
    void GetMeanValueThread( GetMeanValueThreadParams* paramsPtr )
    {
      paramsPtr->m_mutexPtr->lock();
      const unsigned int blockElementsNumber = static_cast<unsigned int>(paramsPtr->m_inputBandPtr->getProperty()->m_blkh *
        paramsPtr->m_inputBandPtr->getProperty()->m_blkw);
      const int blockDataType = paramsPtr->m_inputBandPtr->getProperty()->getType();
      const int blockSizeBytes = paramsPtr->m_inputBandPtr->getBlockSize();
      const double noDataValue = paramsPtr->m_inputBandPtr->getProperty()->m_noDataValue;
      paramsPtr->m_mutexPtr->unlock();

      boost::scoped_array< unsigned char > blockBuffer( new unsigned char[ blockSizeBytes ] );
      boost::scoped_array< double > doubleBuffer( new double[ blockElementsNumber ] );
      unsigned blkX = 0;
      unsigned int elementIdx = 0;
      double mean = 0;
      double meanElementsNumber = 0;
      
      for( unsigned blkY = 0 ; blkY < paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber() ;
        ++blkY )
      {
        for( blkX = 0 ; blkX < paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber() ;
          ++blkX )
        {
          paramsPtr->m_mutexPtr->lock();
          
          if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            paramsPtr->m_mutexPtr->unlock();
          }
          else
          {
            paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            paramsPtr->m_inputBandPtr->read( static_cast<int>(blkX), static_cast<int>(blkY), blockBuffer.get() );
            
            paramsPtr->m_mutexPtr->unlock();
            
            Convert2DoublesVector( blockBuffer.get(), blockDataType, blockElementsNumber,
              doubleBuffer.get() );
              
            for( elementIdx = 0 ; elementIdx < blockElementsNumber ; ++elementIdx )
            {
              if( noDataValue != doubleBuffer[ elementIdx ] )
              {
                mean = 
                  ( 
                    ( 
                      mean 
                      * 
                      meanElementsNumber 
                    ) 
                    + 
                    doubleBuffer[ elementIdx ] 
                  ) 
                  /
                  ( 
                    meanElementsNumber 
                    + 
                    1.0
                  );
                  
                meanElementsNumber = meanElementsNumber + 1.0;
              }
            }
          }
        }
      }
      
      paramsPtr->m_mutexPtr->lock();
      ( *(paramsPtr->m_meanValuePtr) ) = 
        ( 
          ( 
            ( *(paramsPtr->m_meanValuePtr) ) 
            * 
            ( *( paramsPtr->m_pixelsNumberValuePtr ) ) 
          ) 
          + 
          ( 
            mean 
            * 
            meanElementsNumber 
          ) 
        ) 
        /
        ( 
          ( *( paramsPtr->m_pixelsNumberValuePtr ) ) 
          + 
          meanElementsNumber 
        );
      ( *( paramsPtr->m_pixelsNumberValuePtr ) ) += meanElementsNumber;
      paramsPtr->m_mutexPtr->unlock();      
    }
    
    bool GetMeanValue( const te::rst::Band& band, 
      const unsigned int maxThreads, 
      double& meanValue )
    {
      Matrix< bool > rasterBlocksStatus;
      if( ! rasterBlocksStatus.reset( static_cast<unsigned int>(band.getProperty()->m_nblocksy),
        static_cast<unsigned int>(band.getProperty()->m_nblocksx), Matrix< bool >::RAMMemPol ) )
      {
        return false;
      }
      for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
        ++row )
      {
        for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
          ++col )
        {
          rasterBlocksStatus( row, col ) = false;
        }
      }
      
      meanValue = 0.0;
      
      double pixelsNumber = 0.0;
      
      boost::mutex mutex;
      
      GetMeanValueThreadParams params;
      params.m_inputBandPtr = &band;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_meanValuePtr = &meanValue;
      params.m_pixelsNumberValuePtr = &pixelsNumber;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      
      if( maxThreads == 1 )
      {
        GetMeanValueThread( &params );
      }
      else
      {
        unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
        
        boost::thread_group threads;
          
        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads.add_thread( new boost::thread( GetMeanValueThread, 
             &params ) );
        };  
        
        threads.join_all();
      }
      
      return params.m_returnStatus;
    }
    
    void GetStdDevValueThread( GetStdDevValueThreadParams* paramsPtr )
    {
      paramsPtr->m_mutexPtr->lock();
      const unsigned int blockElementsNumber = static_cast<unsigned int>(paramsPtr->m_inputBandPtr->getProperty()->m_blkh *
        paramsPtr->m_inputBandPtr->getProperty()->m_blkw);
      const int blockDataType = paramsPtr->m_inputBandPtr->getProperty()->getType();
      const int blockSizeBytes = paramsPtr->m_inputBandPtr->getBlockSize();
      const double noDataValue = paramsPtr->m_inputBandPtr->getProperty()->m_noDataValue;
      const double& meanValue = paramsPtr->m_meanValue;
      paramsPtr->m_mutexPtr->unlock();

      boost::scoped_array< unsigned char > blockBuffer( new unsigned char[ blockSizeBytes ] );
      boost::scoped_array< double > doubleBuffer( new double[ blockElementsNumber ] );
      unsigned blkX = 0;
      unsigned int elementIdx = 0;
      double diff = 0;
      double squaresDifsSum = 0;
      double elementsNumber = 0;
      
      
      for( unsigned blkY = 0 ; blkY < paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber() ;
        ++blkY )
      {
        for( blkX = 0 ; blkX < paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber() ;
          ++blkX )
        {
          paramsPtr->m_mutexPtr->lock();
          
          if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            paramsPtr->m_mutexPtr->unlock();
          }
          else
          {
            paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            paramsPtr->m_inputBandPtr->read( static_cast<int>(blkX), static_cast<int>(blkY), blockBuffer.get() );
            
            paramsPtr->m_mutexPtr->unlock();
            
            Convert2DoublesVector( blockBuffer.get(), blockDataType, blockElementsNumber,
              doubleBuffer.get() );
              
            for( elementIdx = 0 ; elementIdx < blockElementsNumber ; ++elementIdx )
            {
              if( noDataValue != doubleBuffer[ elementIdx ] )
              {
                diff = doubleBuffer[ elementIdx ] - meanValue;
                diff *= diff;
                squaresDifsSum += diff;
                elementsNumber = elementsNumber + 1.0;
              }
            }
          }
        }
      }
      
      paramsPtr->m_mutexPtr->lock();
      ( *(paramsPtr->m_stdDevValuePtr) ) = 
        ( 
          ( *(paramsPtr->m_stdDevValuePtr) ) 
          + 
          squaresDifsSum 
        );
      ( *( paramsPtr->m_pixelsNumberValuePtr ) ) += elementsNumber;
      paramsPtr->m_mutexPtr->unlock();      
    }
    
    bool GetStdDevValue( const te::rst::Band& band, 
      const unsigned int maxThreads, double const * const meanValuePtr, 
      double& stdDevValue )
    {
      double mean = 0;
      if( meanValuePtr )
      {
        mean = (*meanValuePtr);
      }
      else
      {
        if( ! GetMeanValue( band, maxThreads, mean ) )
        {
          return false;
        }
      }
            
      Matrix< bool > rasterBlocksStatus;
      if( ! rasterBlocksStatus.reset( static_cast<unsigned int>(band.getProperty()->m_nblocksy),
        static_cast<unsigned int>(band.getProperty()->m_nblocksx), Matrix< bool >::RAMMemPol ) )
      {
        return false;
      }
      for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
        ++row )
      {
        for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
          ++col )
        {
          rasterBlocksStatus( row, col ) = false;
        }
      }
      
      stdDevValue = 0.0;
      
      double pixelsNumber = 0.0;
      
      boost::mutex mutex;
      
      GetStdDevValueThreadParams params;
      params.m_inputBandPtr = &band;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_stdDevValuePtr = &stdDevValue;
      params.m_pixelsNumberValuePtr = &pixelsNumber;
      params.m_meanValue = mean;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      
      if( maxThreads == 1 )
      {
        GetStdDevValueThread( &params );
      }
      else
      {
        unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
        
        boost::thread_group threads;
          
        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads.add_thread( new boost::thread( GetStdDevValueThread, 
             &params ) );
        };  
        
        threads.join_all();
      }
      
      stdDevValue = std::sqrt( stdDevValue / pixelsNumber );
      
      return params.m_returnStatus;
    }    
    
    void GetCovarianceValueThread( GetCovarianceValueThreadParams* paramsPtr )
    {
      paramsPtr->m_mutexPtr->lock();
      const unsigned int blockElementsNumber = static_cast<unsigned int>(paramsPtr->m_inputBand1Ptr->getProperty()->m_blkh *
        paramsPtr->m_inputBand1Ptr->getProperty()->m_blkw);
      const int blockDataType1 = paramsPtr->m_inputBand1Ptr->getProperty()->getType();
      const int blockDataType2 = paramsPtr->m_inputBand2Ptr->getProperty()->getType();
      const int blockSizeBytes1 = paramsPtr->m_inputBand1Ptr->getBlockSize();
      const int blockSizeBytes2 = paramsPtr->m_inputBand2Ptr->getBlockSize();
      const double noDataValue1 = paramsPtr->m_inputBand1Ptr->getProperty()->m_noDataValue;
      const double noDataValue2 = paramsPtr->m_inputBand2Ptr->getProperty()->m_noDataValue;
      paramsPtr->m_mutexPtr->unlock();

      boost::scoped_array< unsigned char > blockBuffer1( new unsigned char[ blockSizeBytes1 ] );
      boost::scoped_array< unsigned char > blockBuffer2( new unsigned char[ blockSizeBytes2 ] );
      boost::scoped_array< double > doubleBuffer1( new double[ blockElementsNumber ] );
      boost::scoped_array< double > doubleBuffer2( new double[ blockElementsNumber ] );
      unsigned blkX = 0;
      unsigned int elementIdx = 0;
      double covariance = 0;
      double elementsNumber = 0;
      double diff1 = 0;
      double diff2 = 0;
      
      for( unsigned blkY = 0 ; blkY < paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber() ;
        ++blkY )
      {
        for( blkX = 0 ; blkX < paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber() ;
          ++blkX )
        {
          paramsPtr->m_mutexPtr->lock();
          
          if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            paramsPtr->m_mutexPtr->unlock();
          }
          else
          {
            paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            paramsPtr->m_inputBand1Ptr->read( static_cast<int>(blkX), static_cast<int>(blkY), blockBuffer1.get() );
            paramsPtr->m_inputBand2Ptr->read( static_cast<int>(blkX), static_cast<int>(blkY), blockBuffer2.get() );
            
            paramsPtr->m_mutexPtr->unlock();
            
            Convert2DoublesVector( blockBuffer1.get(), blockDataType1, blockElementsNumber,
              doubleBuffer1.get() );
            Convert2DoublesVector( blockBuffer2.get(), blockDataType2, blockElementsNumber,
              doubleBuffer2.get() );
              
            for( elementIdx = 0 ; elementIdx < blockElementsNumber ; ++elementIdx )
            {
              if( ( noDataValue1 != doubleBuffer1[ elementIdx ] ) &&
                ( noDataValue2 != doubleBuffer2[ elementIdx ] ) )
              {
                diff1 = doubleBuffer1[ elementIdx ] - paramsPtr->m_mean1Value;
                diff2 = doubleBuffer2[ elementIdx ] - paramsPtr->m_mean2Value;
                
                covariance += ( diff1 * diff2 );
                  
                elementsNumber = elementsNumber + 1.0;
              }
            }
          }
        }
      }
      
      paramsPtr->m_mutexPtr->lock();
      ( *(paramsPtr->m_covarianceValuePtr) ) += covariance;
      ( *( paramsPtr->m_pixelsNumberValuePtr ) ) += elementsNumber;
      paramsPtr->m_mutexPtr->unlock();      
    }    
    
    bool GetCovarianceValue( const te::rst::Band& band1, 
      const te::rst::Band& band2, const unsigned int maxThreads, 
      double const * const mean1ValuePtr, double const * const mean2ValuePtr,  
      double& covarianceValue )
    {
      if( ( band1.getProperty()->m_blkw * band1.getProperty()->m_nblocksx ) !=
        ( band2.getProperty()->m_blkw * band2.getProperty()->m_nblocksx ) )
      {
        return false;
      }
      
      if( ( band1.getProperty()->m_blkh * band1.getProperty()->m_nblocksy ) !=
        ( band2.getProperty()->m_blkh * band2.getProperty()->m_nblocksy ) )
      {
        return false;
      }      
      
      double mean1 = 0;
      if( mean1ValuePtr )
      {
        mean1 = (*mean1ValuePtr);
      }
      else
      {
        if( ! GetMeanValue( band1, maxThreads, mean1 ) )
        {
          return false;
        }
      }
      
      double mean2 = 0;
      if( mean2ValuePtr )
      {
        mean2 = (*mean2ValuePtr);
      }
      else
      {
        if( ! GetMeanValue( band2, maxThreads, mean2 ) )
        {
          return false;
        }
      }      
     
      if( ( band1.getProperty()->m_blkw == band2.getProperty()->m_blkw )
        && ( band1.getProperty()->m_blkh == band2.getProperty()->m_blkh )
        && ( band1.getProperty()->m_nblocksx == band2.getProperty()->m_nblocksx )
        && ( band1.getProperty()->m_nblocksy == band2.getProperty()->m_nblocksy ) )
      {
        Matrix< bool > rasterBlocksStatus;
        if( ! rasterBlocksStatus.reset( static_cast<unsigned int>(band1.getProperty()->m_nblocksy),
          static_cast<unsigned int>(band1.getProperty()->m_nblocksx), Matrix< bool >::RAMMemPol ) )
        {
          return false;
        }
        for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
          ++row )
        {
          for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
            ++col )
          {
            rasterBlocksStatus( row, col ) = false;
          }
        }
              
        covarianceValue = 0.0;
        
        double pixelsNumber = 0.0;
        
        boost::mutex mutex;
        
        GetCovarianceValueThreadParams params;
        params.m_inputBand1Ptr = &band1;
        params.m_inputBand2Ptr = &band2;
        params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
        params.m_covarianceValuePtr = &covarianceValue;
        params.m_pixelsNumberValuePtr = &pixelsNumber;
        params.m_returnStatus = true;
        params.m_mutexPtr = &mutex;
        params.m_mean1Value = mean1;
        params.m_mean2Value = mean2;
        
        if( maxThreads == 1 )
        {
          GetCovarianceValueThread( &params );
        }
        else
        {
          unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
          
          boost::thread_group threads;
            
          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            threads.add_thread( new boost::thread( GetCovarianceValueThread, 
               &params ) );
          };  
          
          threads.join_all();
        }
        
        if( pixelsNumber != 0.0 )
        {
          covarianceValue /= pixelsNumber;
        }
        
        return params.m_returnStatus;
      }
      else
      {
        const unsigned int nCols = static_cast<unsigned int>(band1.getProperty()->m_blkw * band1.getProperty()->m_nblocksx);
        const unsigned int nRows = static_cast<unsigned int>(band1.getProperty()->m_blkh * band1.getProperty()->m_nblocksy);
        double value1 = 0;
        double value2 = 0;
        const double noDataValue1 = band1.getProperty()->m_noDataValue;
        const double noDataValue2 = band2.getProperty()->m_noDataValue;    
        unsigned int col = 0;
        double pixelsNumber = 0.0;
        
        covarianceValue = 0;
        
        for( unsigned int row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            band1.getValue( col, row, value1 );
            band2.getValue( col, row, value2 );
            
            if( ( noDataValue1 != value1 ) &&
              ( noDataValue2 != value2 ) )
            {
              covarianceValue += ( ( value1 - mean1 ) * ( value2 - mean2 ) );
                
              pixelsNumber = pixelsNumber + 1.0;
            }            
          }
        }
        
        if( pixelsNumber != 0.0 )
        {
          covarianceValue /= pixelsNumber;
        }
        
        return true;
      }
    }    
    
    bool DirectPrincipalComponents( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      boost::numeric::ublas::matrix< double >& pcaMatrix,
      te::rst::Raster& pcaRaster,
      const std::vector< unsigned int >& pcaRasterBands,
      const unsigned int maxThreads )
    {
      if( ( inputRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
      {
        return false;
      }
      if( ( pcaRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }   
      if( inputRaster.getNumberOfBands() != pcaRaster.getNumberOfBands() )
      {
        return false;
      }
      if( pcaRaster.getNumberOfBands() != inputRasterBands.size() )
      {
        return false;
      }
      if( inputRaster.getNumberOfRows() != pcaRaster.getNumberOfRows() )
      {
        return false;
      }
      if( inputRaster.getNumberOfColumns() != pcaRaster.getNumberOfColumns() )
      {
        return false;
      }    
      
      // Covariance matrix
      
      boost::numeric::ublas::matrix< double > covarianceMatrix;
      
      covarianceMatrix.resize( inputRasterBands.size(), inputRasterBands.size() );
      
      for( unsigned int covMatrixIdx1 = 0 ; covMatrixIdx1 < inputRasterBands.size() ;
        ++covMatrixIdx1 )
      {
        if( inputRasterBands[ covMatrixIdx1 ] >= inputRaster.getNumberOfBands() )
        {
          return false;
        }
        
        for( unsigned int covMatrixIdx2 = 0 ; covMatrixIdx2 < inputRasterBands.size() ;
        ++covMatrixIdx2 )
        {
          if( inputRasterBands[ covMatrixIdx2 ] >= inputRaster.getNumberOfBands() )
          {
            return false;
          }    
          
          if( covMatrixIdx1 > covMatrixIdx2 )
          {
            covarianceMatrix( covMatrixIdx1, covMatrixIdx2 ) =
              covarianceMatrix( covMatrixIdx2, covMatrixIdx1 );
          }
          else
          {
            if( ! GetCovarianceValue( *( inputRaster.getBand( inputRasterBands[ covMatrixIdx1 ] ) ),
               *( inputRaster.getBand( inputRasterBands[ covMatrixIdx2 ] ) ),
               maxThreads, 0,  0, covarianceMatrix( covMatrixIdx1, covMatrixIdx2 ) ) )
            {
              return false;
            }
          }
        }
      }
      
//std::cout << std::endl << "Covariance matrix:" << covarianceMatrix << std::endl;      
      
      // Eigen stuff
      
      boost::numeric::ublas::matrix< double > eigenValues;
      boost::numeric::ublas::matrix< double > eigenVectors;
      
      if( ! te::common::EigenVectors( covarianceMatrix, eigenVectors, eigenValues ) )
      {
        return false;
      }
      
      pcaMatrix = boost::numeric::ublas::trans( eigenVectors );
      
      return RemapValues( inputRaster, inputRasterBands, pcaMatrix, pcaRaster, 
        pcaRasterBands, maxThreads );
    }
    
    bool InversePrincipalComponents( 
      const te::rst::Raster& pcaRaster,
      const boost::numeric::ublas::matrix< double >& pcaMatrix,
      te::rst::Raster& outputRaster,
      const std::vector< unsigned int >& outputRasterBands,
      const unsigned int maxThreads )    
    {
      boost::numeric::ublas::matrix< double > inversePcaMatrix;
      if( ! te::common::GetInverseMatrix( pcaMatrix, inversePcaMatrix ) )
      {
        return false;
      }
      
      std::vector< unsigned int > inputRasterBands;
      
      for( unsigned int bandIdx = 0 ; bandIdx < pcaRaster.getNumberOfBands() ;
        ++bandIdx )
      {
        inputRasterBands.push_back( bandIdx );
      }
      
      return RemapValues( pcaRaster, inputRasterBands, inversePcaMatrix, 
        outputRaster, outputRasterBands, maxThreads );
    }
    
    void RemapValuesThread( RemapValuesThreadParams* paramsPtr )
    {
      paramsPtr->m_mutexPtr->lock();
      
      assert( paramsPtr->m_inputRasterBandsPtr->operator[]( 0 ) < 
        paramsPtr->m_inputRasterPtr->getNumberOfBands() );
      const unsigned int blockElementsNumber = static_cast<unsigned int>(paramsPtr->m_inputRasterPtr->getBand(
        paramsPtr->m_inputRasterBandsPtr->operator[]( 0 ) )->getProperty()->m_blkh *
        paramsPtr->m_inputRasterPtr->getBand( paramsPtr->m_inputRasterBandsPtr->operator[]( 0 ) )->getProperty()->m_blkw);
      
      const std::vector< unsigned int >  inputRasterBands = *( paramsPtr->m_inputRasterBandsPtr );
      const unsigned int  inputRasterBandsSize = static_cast<unsigned int>(inputRasterBands.size());
      const std::vector< unsigned int >  outputRasterBands = *( paramsPtr->m_inputRasterBandsPtr );
      assert( inputRasterBandsSize == paramsPtr->m_outputRasterPtr->getNumberOfBands() );
      assert( inputRasterBandsSize == outputRasterBands.size() );
      
      unsigned int maxBlocksSizesBytes = 0;
      std::vector< double > inputBandsNoDataValues( inputRasterBandsSize, 0.0 );
      std::vector< double > outputBandsNoDataValues( inputRasterBandsSize, 0.0 );
      std::vector< boost::shared_array< double > > inDoubleBuffersHandlers( inputRasterBandsSize );
      std::vector< boost::shared_array< double > > outDoubleBuffersHandlers( inputRasterBandsSize );
      unsigned int inputRasterBandsIdx = 0;
      boost::numeric::ublas::matrix< double > remapMatrix = *( paramsPtr->m_remapMatrixPtr );
      std::vector< double > outputMinValue( inputRasterBandsSize );
      std::vector< double > outputMaxValue( inputRasterBandsSize );        
      boost::shared_array< double* > inDoubleBuffersPtrsHandler( new double*[ inputRasterBandsSize ] );
      boost::shared_array< double* > outDoubleBuffersPtrsHandler( new double*[ inputRasterBandsSize ] );
      double** inDoubleBuffers = inDoubleBuffersPtrsHandler.get();
      double** outDoubleBuffers = outDoubleBuffersPtrsHandler.get();
      
      for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
        ++inputRasterBandsIdx )
      {
        const unsigned int& inBandIdx = inputRasterBands[ inputRasterBandsIdx ];
        assert( inBandIdx < paramsPtr->m_inputRasterPtr->getNumberOfBands() );
        
        te::rst::Band const * const inBandPtr = paramsPtr->m_inputRasterPtr->getBand( inBandIdx );
        te::rst::Band * const outBandPtr = paramsPtr->m_outputRasterPtr->getBand( 
          outputRasterBands[ inputRasterBandsIdx ] );
        
        maxBlocksSizesBytes = std::max( maxBlocksSizesBytes, static_cast<unsigned int>(inBandPtr->getBlockSize()) );
        maxBlocksSizesBytes = std::max( maxBlocksSizesBytes, static_cast<unsigned int>(outBandPtr->getBlockSize()) );
        
        inputBandsNoDataValues[ inputRasterBandsIdx ] = inBandPtr->getProperty()->m_noDataValue;
        outputBandsNoDataValues[ inputRasterBandsIdx ] = outBandPtr->getProperty()->m_noDataValue;
          
        inDoubleBuffersHandlers[ inputRasterBandsIdx ].reset( new double[ blockElementsNumber ] );
        inDoubleBuffers[ inputRasterBandsIdx ] = inDoubleBuffersHandlers[ inputRasterBandsIdx ].get();
        
        outDoubleBuffersHandlers[ inputRasterBandsIdx ].reset( new double[ blockElementsNumber ] );
        outDoubleBuffers[ inputRasterBandsIdx ] = outDoubleBuffersHandlers[ inputRasterBandsIdx ].get();
        
        te::rst::GetDataTypeRanges( 
          outBandPtr->getProperty()->getType(),
          outputMinValue[ inputRasterBandsIdx ],
          outputMaxValue[ inputRasterBandsIdx ] );        
      }
      
      paramsPtr->m_mutexPtr->unlock();

      boost::scoped_array< unsigned char > blockBuffer( new unsigned char[ maxBlocksSizesBytes ] );
      unsigned blkX = 0;
      unsigned int elementIdx = 0;
      boost::numeric::ublas::matrix< double > pixelValues( inputRasterBandsSize, 1 );
      boost::numeric::ublas::matrix< double > remappedPixelValues( inputRasterBandsSize, 1 );
      bool elementIsValid = false;
      
      for( unsigned blkY = 0 ; blkY < paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber() ;
        ++blkY )
      {
        for( blkX = 0 ; blkX < paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber() ;
          ++blkX )
        {
          paramsPtr->m_mutexPtr->lock();
          
          if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            paramsPtr->m_mutexPtr->unlock();
          }
          else
          {
            paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            // reading the input blocks
            
            for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
              ++inputRasterBandsIdx )
            {    
              const unsigned int& inBandIdx = inputRasterBands[ inputRasterBandsIdx ];
              te::rst::Band const * const inBandPtr = paramsPtr->m_inputRasterPtr->getBand( inBandIdx );
              
              inBandPtr->read( static_cast<int>(blkX), static_cast<int>(blkY), blockBuffer.get() );
              Convert2DoublesVector( blockBuffer.get(), 
                inBandPtr->getProperty()->getType(), 
                blockElementsNumber,
                inDoubleBuffers[ inputRasterBandsIdx ] );            
            }
            
            paramsPtr->m_mutexPtr->unlock();
            
            // Remapping the values using the remap matrix
              
            for( elementIdx = 0 ; elementIdx < blockElementsNumber ; ++elementIdx )
            {
              elementIsValid = true;
              
              for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
                ++inputRasterBandsIdx )
              {              
                if( inDoubleBuffers[ inputRasterBandsIdx ][ elementIdx ] == 
                  inputBandsNoDataValues[ inputRasterBandsIdx ] )
                {
                  elementIsValid = false;
                  break;
                }
                else
                {
                  pixelValues( inputRasterBandsIdx, 0 ) = 
                    inDoubleBuffers[ inputRasterBandsIdx ][ elementIdx ];
                }
              }              
              
              if( elementIsValid )
              {
//std::cout << std::endl << "pixelValues:" << pixelValues << std::endl;
                
                remappedPixelValues = boost::numeric::ublas::prod( remapMatrix, pixelValues );
                
//std::cout << std::endl << "remappedPixelValues:" << remappedPixelValues << std::endl;
                
                for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
                  ++inputRasterBandsIdx )
                {              
                  outDoubleBuffers[ inputRasterBandsIdx ][ elementIdx ] =
                    std::max(
                      outputMinValue[ inputRasterBandsIdx ],
                      std::min(
                        outputMaxValue[ inputRasterBandsIdx ],
                        remappedPixelValues( inputRasterBandsIdx, 0 )
                      )
                    );
                } 
              }
              else
              {
                for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
                  ++inputRasterBandsIdx )
                {              
                  outDoubleBuffers[ inputRasterBandsIdx ][ elementIdx ] =
                    outputBandsNoDataValues[ inputRasterBandsIdx ] ;
                } 
              }
            }
            
            // Writing the remmaped blocks
            
            paramsPtr->m_mutexPtr->lock();
            
            for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
              ++inputRasterBandsIdx )
            { 
              te::rst::Band* outBandPtr = paramsPtr->m_outputRasterPtr->getBand( 
                outputRasterBands[ inputRasterBandsIdx ] );
              
              ConvertDoublesVector( outDoubleBuffers[ inputRasterBandsIdx ],
                blockElementsNumber, outBandPtr->getProperty()->getType(),
                blockBuffer.get() );              
              
              outBandPtr->write( static_cast<int>(blkX), static_cast<int>(blkY), blockBuffer.get() );
            }            
            
            paramsPtr->m_mutexPtr->unlock();
          }
        }
      }
     
    }      

    bool RemapValues( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands, 
      const boost::numeric::ublas::matrix< double >& remapMatrix,
      te::rst::Raster& outputRaster,
      const std::vector< unsigned int >& outputRasterBands,
      const unsigned int maxThreads )
    {
      if( ( inputRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
      {
        return false;
      }
      if( ( outputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }   
      if( remapMatrix.size1() != inputRasterBands.size() )
      {
        return false;
      }
      if( remapMatrix.size2() != inputRasterBands.size() )
      {
        return false;
      }      
      if( outputRaster.getNumberOfBands() != inputRasterBands.size() )
      {
        return false;
      }
      if( inputRaster.getNumberOfRows() != outputRaster.getNumberOfRows() )
      {
        return false;
      }
      if( inputRaster.getNumberOfColumns() != outputRaster.getNumberOfColumns() )
      {
        return false;
      }    
      if( remapMatrix.size1() != outputRasterBands.size() )
      {
        return false;
      }
      if( remapMatrix.size2() != outputRasterBands.size() )
      {
        return false;
      }     
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBands.size() ;
        ++inputRasterBandsIdx )
      {
        if( inputRasterBands[ inputRasterBandsIdx ] >= inputRaster.getNumberOfBands() )
        {
          return false;
        }    
      }
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx < outputRasterBands.size() ;
        ++outputRasterBandsIdx )
      {
        if( outputRasterBands[ outputRasterBandsIdx ] >= outputRaster.getNumberOfBands() )
        {
          return false;
        }    
      }      
      
      // Checking if optmized PCA can be executed
      
      bool useOptimizedRemap = true;
      
      {
        for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx <
          inputRaster.getNumberOfBands() ; ++inputRasterBandsIdx )
        {
          if(
              ( 
                inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_blkw 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_blkw 
              )
              ||
              ( 
                inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_blkh 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_blkh
              )              
              ||
              ( 
                inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_nblocksx 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_nblocksx 
              )    
              ||
              ( 
                inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_nblocksy 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_nblocksy 
              )  
              ||
              ( 
                outputRaster.getBand( inputRasterBandsIdx )->getProperty()->m_blkw 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_blkw 
              )
              ||
              ( 
                outputRaster.getBand( inputRasterBandsIdx )->getProperty()->m_blkh 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_blkh
              )              
              ||
              ( 
                outputRaster.getBand( inputRasterBandsIdx )->getProperty()->m_nblocksx 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_nblocksx 
              )    
              ||
              ( 
                outputRaster.getBand( inputRasterBandsIdx )->getProperty()->m_nblocksy 
                != 
                inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_nblocksy 
              )                                     
            )
          {
            useOptimizedRemap = false;
            break;
          }
        }
      }
      
      // remapping
      
      if( useOptimizedRemap )
      {
        Matrix< bool > rasterBlocksStatus;
        if( ! rasterBlocksStatus.reset( 
          static_cast<unsigned int>(inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_nblocksy),
          static_cast<unsigned int>(inputRaster.getBand( inputRasterBands[ 0 ] )->getProperty()->m_nblocksx),
          Matrix< bool >::RAMMemPol ) )
        {
          return false;
        }
        for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
          ++row )
        {
          for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
            ++col )
          {
            rasterBlocksStatus( row, col ) = false;
          }
        }
        
        boost::mutex mutex;
        
        RemapValuesThreadParams params;
        params.m_inputRasterPtr = &inputRaster;
        params.m_inputRasterBandsPtr = &inputRasterBands;
        params.m_outputRasterPtr = &outputRaster;
        params.m_outputRasterBandsPtr = &outputRasterBands;
        params.m_remapMatrixPtr = &remapMatrix;
        params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
        params.m_returnStatus = true;
        params.m_mutexPtr = &mutex;
        
        if( maxThreads == 1 )
        {
          RemapValuesThread( &params );
        }
        else
        {
          unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
          
          boost::thread_group threads;
            
          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            threads.add_thread( new boost::thread( RemapValuesThread, 
               &params ) );
          };  
          
          threads.join_all();
        }
        
        return params.m_returnStatus;        
      }
      else
      {
        const unsigned int inputRasterBandsSize = static_cast<unsigned int>(inputRasterBands.size());
        const unsigned int nRows = inputRaster.getNumberOfRows();
        const unsigned int nCols = inputRaster.getNumberOfColumns();
        boost::numeric::ublas::matrix< double > pixelValues( inputRasterBands.size(), 1 );
        boost::numeric::ublas::matrix< double > remappedPixelValues( inputRasterBands.size(), 1 );
        std::vector< double > inputNoDataValues( inputRasterBandsSize );
        std::vector< double > outputNoDataValues( inputRasterBandsSize );
        std::vector< double > outputMinValue( inputRasterBandsSize );
        std::vector< double > outputMaxValue( inputRasterBandsSize );        
        unsigned int inputRasterBandsIdx = 0;
        unsigned int row = 0;
        unsigned int col = 0;
        bool pixelIsValid = false;
        
        for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
          ++inputRasterBandsIdx )
        {
          if( inputRasterBands[ inputRasterBandsIdx ] >= inputRaster.getNumberOfBands() )
          {
            return false;
          }
          
          inputNoDataValues[ inputRasterBandsIdx ] = inputRaster.getBand( 
            inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_noDataValue;
            
          outputNoDataValues[ inputRasterBandsIdx ] = outputRaster.getBand( 
            outputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_noDataValue;
            
          te::rst::GetDataTypeRanges( 
            outputRaster.getBand( outputRasterBands[ inputRasterBandsIdx ] )->getProperty()->getType(),
            outputMinValue[ inputRasterBandsIdx ],
            outputMaxValue[ inputRasterBandsIdx ] );
        }
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            pixelIsValid = true;
            
            for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
              ++inputRasterBandsIdx )
            {
              inputRaster.getValue( col, row, pixelValues( inputRasterBandsIdx, 0 ),
                inputRasterBands[ inputRasterBandsIdx ] );
              
              if( pixelValues( inputRasterBandsIdx, 0 ) == inputNoDataValues[ inputRasterBandsIdx ] )
              {
                pixelIsValid = false;
                break;
              }
            }
            
            if( pixelIsValid )
            {
              remappedPixelValues = boost::numeric::ublas::prod( remapMatrix, pixelValues );
              
              for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
                ++inputRasterBandsIdx )
              {              
                outputRaster.setValue( 
                  col, 
                  row, 
                  std::max(
                    outputMinValue[ inputRasterBandsIdx ]
                    ,
                    std::min( 
                      outputMaxValue[ inputRasterBandsIdx ]
                      ,
                      remappedPixelValues( inputRasterBandsIdx, 0 )
                    )
                  ),
                  outputRasterBands[ inputRasterBandsIdx ] );
              }              
            }
            else
            {
              for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBandsSize ;
                ++inputRasterBandsIdx )
              {              
                outputRaster.setValue( col, row, outputNoDataValues[ inputRasterBandsIdx ], 
                  outputRasterBands[ inputRasterBandsIdx ] );
              }
            }
          }
        }
      }
        
      return true;
    }
    
    bool DecomposeBands( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      const std::vector< std::map<std::string, std::string> > & outputRastersInfos,
      const std::string& outputDataSourceType,
      std::vector< boost::shared_ptr< te::rst::Raster > > & outputRastersPtrs )
    {
      outputRastersPtrs.clear();
      
      if( !( inputRaster.getAccessPolicy() & te::common::RAccess ) )
      {
        return false;
      }
      if( outputRastersInfos.size() != inputRasterBands.size() )
      {
        return false;
      }
      if( outputDataSourceType.empty() )
      {
        return false;
      }
      
      outputRastersPtrs.resize( inputRasterBands.size() );
      
      const unsigned int nRows = inputRaster.getNumberOfRows();
      const unsigned int nCols = inputRaster.getNumberOfColumns();
      
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx <
        inputRasterBands.size() ; ++inputRasterBandsIdx )
      {
        const unsigned int bandIdx = inputRasterBands[ inputRasterBandsIdx ];
        if( bandIdx >= inputRaster.getNumberOfBands() )
        {
          return false;
        }
        
        std::vector< te::rst::BandProperty* > bandsProperties;

        bandsProperties.push_back( new te::rst::BandProperty( 
          *( inputRaster.getBand( bandIdx )->getProperty()) ) );
        
        outputRastersPtrs[ inputRasterBandsIdx].reset( 
          te::rst::RasterFactory::make( outputDataSourceType,
          new te::rst::Grid( *inputRaster.getGrid() ), bandsProperties, 
          outputRastersInfos[ inputRasterBandsIdx ], 0, 0 ) );
        if( outputRastersPtrs[ inputRasterBandsIdx].get() == 0 ) return false;        
        
        unsigned int col = 0;
        unsigned int row = 0;
        double value = 0;
        const te::rst::Band& inBand = *(inputRaster.getBand( bandIdx ));
        te::rst::Band& outBand = *(outputRastersPtrs[ inputRasterBandsIdx]->getBand( 0 ));
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, value );
            outBand.setValue( col, row, value );
          }
        }
      }
      
      return true;
    }
    
    bool ComposeBands( 
      te::rp::FeederConstRaster& feeder,
      const std::vector< unsigned int >& inputRasterBands,
      const te::rst::Interpolator::Method& interpMethod,
      const std::map<std::string, std::string>& outputRasterInfo,
      const std::string& outputDataSourceType,
      std::unique_ptr< te::rst::Raster >& outputRasterPtr )
    {
      outputRasterPtr.reset();
      
      if( inputRasterBands.size() != feeder.getObjsCount() )
      {
        return false;
      }
      if( outputDataSourceType.empty() )
      {
        return false;
      }
      
      // creating the output raster
      
      {
        te::rst::Raster const * inputRasterPtr = 0;
        std::unique_ptr< te::rst::Grid > outputGridPtr;
        std::vector< te::rst::BandProperty* > bandsProperties;
        
        feeder.reset();
        while( (inputRasterPtr = feeder.getCurrentObj()) )
        {
          if( inputRasterBands[ feeder.getCurrentOffset() ] >= 
            inputRasterPtr->getNumberOfBands() )
          {
            return false;
          }
          
          if( feeder.getCurrentOffset() == 0 )
          {
            outputGridPtr.reset( new te::rst::Grid( *inputRasterPtr->getGrid() ) );
          }          

          bandsProperties.push_back( new te::rst::BandProperty( 
            *( inputRasterPtr->getBand( inputRasterBands[ 
            feeder.getCurrentOffset() ] )->getProperty()) ) );
          
          feeder.moveNext();
        }
        
        outputRasterPtr.reset( te::rst::RasterFactory::make( outputDataSourceType,
          outputGridPtr.release(), bandsProperties, outputRasterInfo, 0, 0 ) );
        if( outputRasterPtr.get() == 0 ) return false;
      }
      
      // copying data from each input band
        
      {
        te::rst::Raster const * inputRasterPtr = 0;
        
        feeder.reset();
        while( (inputRasterPtr = feeder.getCurrentObj()) )
        {
          const unsigned int inBandIdx = inputRasterBands[ 
            feeder.getCurrentOffset() ];
          te::rst::Interpolator interp( inputRasterPtr, interpMethod );
          unsigned int outRow = 0;
          unsigned int outCol = 0;
          const unsigned int nOutRows = outputRasterPtr->getNumberOfRows();
          const unsigned int nOutCols = outputRasterPtr->getNumberOfColumns();
          te::rst::Band& outBand = *outputRasterPtr->getBand( 
            feeder.getCurrentOffset() );
          const te::rst::Grid& inGrid = *inputRasterPtr->getGrid();
          const te::rst::Grid& outGrid = *outputRasterPtr->getGrid();
          double xOutCoord = 0;
          double yOutCoord = 0;
          double xInCoord = 0;
          double yInCoord = 0;          
          double inRow = 0;
          double inCol = 0;
          std::complex< double > value = 0;
          te::srs::Converter conv( outputRasterPtr->getSRID(),
            inputRasterPtr->getSRID() );
          
          if( inputRasterPtr->getSRID() == outputRasterPtr->getSRID() )
          {
            for( outRow = 0 ; outRow < nOutRows ; ++outRow )
            {
              for( outCol = 0 ; outCol < nOutCols ; ++outCol )
              {
                outGrid.gridToGeo( (double)outCol, (double)outRow, xOutCoord, yOutCoord );
                inGrid.geoToGrid( xOutCoord, yOutCoord, inCol, inRow );
                interp.getValue( inCol, inRow, value, inBandIdx );
                outBand.setValue( outCol, outRow, value );
              }
            }
          }
          else
          {
            for( outRow = 0 ; outRow < nOutRows ; ++outRow )
            {
              for( outCol = 0 ; outCol < nOutCols ; ++outCol )
              {
                outGrid.gridToGeo( (double)outCol, (double)outRow, xOutCoord, yOutCoord );
                conv.convert( xOutCoord, yOutCoord, xInCoord, yInCoord );
                inGrid.geoToGrid( xInCoord, yInCoord, inCol, inRow );
                interp.getValue( inCol, inRow, value, inBandIdx );
                outBand.setValue( outCol, outRow, value );                
              }
            }
          }
          
          feeder.moveNext();
        }
      }        
      
      return true;
    }
    
    bool GetDetailedExtent( const te::rst::Grid& grid, te::gm::LinearRing& detailedExtent )
    {
      const unsigned int nCols = grid.getNumberOfColumns();
      const unsigned int nRows = grid.getNumberOfRows();
      if( ( nCols == 0 ) || ( nRows == 0 ) )
      {
        return false;
      }
      
      const unsigned int ringSize = ( 2 * ( nCols + 1 ) ) +
        ( 2 * ( nRows - 1 ) ) + 1;
      
      te::gm::LinearRing ring( ringSize , te::gm::LineStringType,
        grid.getSRID(), 0 );

      const double lLX = grid.getExtent()->getLowerLeftX();
      const double lLY = grid.getExtent()->getLowerLeftY();
      const double uRX = grid.getExtent()->getUpperRightX();
      const double uRY = grid.getExtent()->getUpperRightY();
      const double& resX = grid.getResolutionX();
      const double& resY = grid.getResolutionY();        
      unsigned int ringIdx = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      
      ring.setPoint( 0, lLX, uRY ); 
      
      for( col = 1 ; col <= nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, lLX + ( (static_cast<double>( col ) ) * resX ), uRY );
      }

      for( row = 1 ; row <= nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, uRX, uRY - ( (static_cast<double>( row ) ) * resY ) );
      } 
      
      for( col = 1 ; col <= nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, uRX - ( (static_cast<double>( col ) ) * resX ), lLY );
      }    
      
      for( row = 1 ; row <= nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, lLX, lLY + ( (static_cast<double>( row ) ) * resY ) );
      }

      // Ensuring correct closing
      ring.setPoint( ringSize - 1, lLX, uRY );
      
      detailedExtent = ring;
      
      return true;
    }
    
    bool GetIndexedDetailedExtent( const te::rst::Grid& grid, 
      te::gm::LinearRing& indexedDetailedExtent )
    {
      const unsigned int nCols = grid.getNumberOfColumns();
      const unsigned int nRows = grid.getNumberOfRows();
      if( ( nCols == 0 ) || ( nRows == 0 ) )
      {
        return false;
      }
      
      const unsigned int ringSize = ( 2 * ( nCols + 1 ) ) +
        ( 2 * ( nRows - 1 ) ) + 1;
      
      te::gm::LinearRing ring( ringSize , te::gm::LineStringType,
        0, 0 );

      const double lLY = (static_cast<double>(nRows)) - 0.5;
      const double uRX = (static_cast<double>(nCols)) - 0.5;
      unsigned int ringIdx = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      
      ring.setPoint( 0, -0.5, -0.5 );
      
      for( col = 0 ; col < nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, 0.5 + (static_cast<double>(col)), (-0.5) );
      }

      for( row = 0 ; row < nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, uRX, 0.5 + (static_cast<double>(row)) );
      } 
      
      for( col = 1; col <= nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, uRX - (static_cast<double>(col)), lLY );
      }    
      
      for( row = 1 ; row <= nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, (-0.5), lLY - (static_cast<double>(row)) );
      }       

      // Ensuring correct closing
      ring.setPoint( ringSize - 1, -0.5, -0.5 ); 
      
      indexedDetailedExtent = ring;
      
      return true;
    }
    
    boost::numeric::ublas::matrix< double > CreateWaveletAtrousFilter( 
      const WaveletAtrousFilterType& filterType )
    {
      boost::numeric::ublas::matrix< double > emptyFilter;
      
      switch( filterType )
      {
        case B3SplineWAFilter :
        {
          boost::numeric::ublas::matrix< double > internalFilter( 5, 5 );
          const double weight = 256;
          
          internalFilter(0,0) = 1/weight; internalFilter(0,1) = 4/weight; internalFilter(0,2) = 6/weight; internalFilter(0,3) = 4/weight; internalFilter(0,4) = 1/weight;
          internalFilter(1,0) = 4/weight; internalFilter(1,1) = 16/weight; internalFilter(1,2) = 24/weight; internalFilter(1,3) = 16/weight; internalFilter(1,4) = 4/weight;
          internalFilter(2,0) = 6/weight; internalFilter(2,1) = 24/weight; internalFilter(2,2) = 36/weight; internalFilter(2,3) = 24/weight; internalFilter(2,4) = 6/weight;
          internalFilter(3,0) = 4/weight; internalFilter(3,1) = 16/weight; internalFilter(3,2) = 24/weight; internalFilter(3,3) = 16/weight; internalFilter(3,4) = 4/weight;
          internalFilter(4,0) = 1/weight; internalFilter(4,1) = 4/weight; internalFilter(4,2) = 6/weight; internalFilter(4,3) = 4/weight; internalFilter(4,4) = 1/weight;
          
          return internalFilter;
        }
        case TriangleWAFilter :
        {
          boost::numeric::ublas::matrix< double > internalFilter( 3, 3 );
          const double weight = 16;
          
          internalFilter(0,0) = 1/weight; internalFilter(0,1) = 2/weight; internalFilter(0,2) = 1/weight;
          internalFilter(1,0) = 2/weight; internalFilter(1,1) = 4/weight; internalFilter(1,2) = 2/weight;
          internalFilter(2,0) = 1/weight; internalFilter(2,1) = 2/weight; internalFilter(2,2) = 1/weight;
          
          return internalFilter;
        }  
        default :
        {
          throw te::rp::Exception( "Invalid filter type" );
        }
      }
      
      return emptyFilter;
    }
    
    bool DirectWaveletAtrous( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      te::rst::Raster& waveletRaster,
      const unsigned int levelsNumber,
      const boost::numeric::ublas::matrix< double >& filter )
    {
      if( ! ( inputRaster.getAccessPolicy() & te::common::RAccess ) )
      {
        return false;
      }
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx <
        inputRasterBands.size() ; ++inputRasterBandsIdx )
      {
        if( inputRasterBands[ inputRasterBandsIdx ] >= inputRaster.getNumberOfBands() )
        {
          return false;
        }
      }
      if( 
          ( ( waveletRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
          ||
          ( waveletRaster.getNumberOfColumns() != inputRaster.getNumberOfColumns() )
          ||
          ( waveletRaster.getNumberOfRows() != inputRaster.getNumberOfRows() )
          ||
          ( waveletRaster.getNumberOfBands() < ( 2 *  levelsNumber * inputRasterBands.size() ) )
        )
      {
        return false;
      }
      if( levelsNumber == 0 )
      {
        return false;
      }
      if( ( filter.size1() == 0 ) || ( filter.size2() == 0 ) || 
        ( filter.size1() != filter.size2() )
      )
      {
        return false;
      }
      
      // Creating the coeficients and the resitual planes for each required
      // wavelet level

      const int filterWidth = static_cast<int>(filter.size1());
      const int offset = filterWidth / 2;
      const int nLines = static_cast<int>(inputRaster.getNumberOfRows());
      const int nCols = static_cast<int>(inputRaster.getNumberOfColumns());
      
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBands.size() ;
        ++inputRasterBandsIdx )
      {
        for( unsigned int levelIndex = 0; levelIndex < levelsNumber; ++levelIndex)
        {
          const unsigned int currentSmoothBandIdx = ( 2 * levelsNumber * 
            inputRasterBandsIdx ) + ( 2 * levelIndex );
          te::rst::Band& currentSmoothBand = *waveletRaster.getBand( currentSmoothBandIdx );
          
          const unsigned int currentWaveletBandIdx = currentSmoothBandIdx + 1;
          te::rst::Band& currentWaveletBand = *waveletRaster.getBand(
            currentWaveletBandIdx );
          
          const unsigned int prevSmoothBandIdx =  currentSmoothBandIdx - 2;
          te::rst::Band const* prevSmoothBandPtr = 0;
          if( levelIndex == 0 )
          {
            prevSmoothBandPtr = inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] );
          }
          else
          {
            prevSmoothBandPtr = waveletRaster.getBand( prevSmoothBandIdx );
          }          
          const te::rst::Band& prevSmoothBand = *prevSmoothBandPtr;          
          
          const int filterScale = static_cast<int>(std::pow(2.0, static_cast<double>(levelIndex)));
          
          int col = 0;
          int row = 0;
          int convolutionCenterCol = 0;
          int convolutionCenterRow = 0;
          int filterCol = 0;
          int filterRow = 0; 
          double valueOriginal = 0.0;
          double valueNew = 0.0;          

          for (convolutionCenterRow = 0; convolutionCenterRow < nLines; convolutionCenterRow++)
          {
            for (convolutionCenterCol = 0; convolutionCenterCol < nCols; convolutionCenterCol++)
            {
              valueNew = 0.0;
              
              for (filterRow = 0; filterRow < filterWidth; filterRow++)
              {
                for (filterCol = 0; filterCol < filterWidth; filterCol++)
                {
                  col = convolutionCenterCol+(filterCol-offset)*filterScale;
                  row = convolutionCenterRow+(filterRow-offset)*filterScale;
                  
                  if (col < 0)
                    col = nCols + col;
                  else if (col >= nCols)
                    col = col - nCols;
                  if (row < 0)
                    row = nLines + row;
                  else if (row >= nLines)
                    row = row - nLines;
                  
                  prevSmoothBand.getValue( static_cast<unsigned int>(col), static_cast<unsigned int>(row), valueOriginal );
                  
                  valueNew += valueOriginal * filter( static_cast<size_t>(filterRow), static_cast<size_t>(filterCol) );
                }
              }
              
              currentSmoothBand.setValue( static_cast<unsigned int>(convolutionCenterCol), static_cast<unsigned int>(convolutionCenterRow),
                valueNew );
              prevSmoothBand.getValue( static_cast<unsigned int>(convolutionCenterCol), static_cast<unsigned int>(convolutionCenterRow),
                valueOriginal );
              currentWaveletBand.setValue( static_cast<unsigned int>(convolutionCenterCol), static_cast<unsigned int>(convolutionCenterRow),
                valueOriginal - valueNew );
            }
          }
        }
      }
      
      return true;
    }
    
    bool InverseWaveletAtrous( 
      const te::rst::Raster& waveletRaster,
      const unsigned int levelsNumber,
      te::rst::Raster& outputRaster,
      const std::vector< unsigned int >& outputRasterBands )
    {
      if( 
          ( ! ( waveletRaster.getAccessPolicy() & te::common::RAccess ) )
          ||
          ( waveletRaster.getNumberOfBands() < ( 2 *  levelsNumber * outputRasterBands.size() ) )
        )
      {
        return false;
      }
      if( levelsNumber == 0 )
      {
        return false;
      }
      if( 
          ( ( outputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
          ||
          ( waveletRaster.getNumberOfColumns() != outputRaster.getNumberOfColumns() )
          ||
          ( waveletRaster.getNumberOfRows() != outputRaster.getNumberOfRows() )
        )
      {
        return false;
      }
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterBands.size() ; ++outputRasterBandsIdx )
      {
        if( outputRasterBands[ outputRasterBandsIdx ] >= outputRaster.getNumberOfBands() )
        {
          return false;
        }
      }
      
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterBands.size() ; ++outputRasterBandsIdx )
      {
        const unsigned int outputRasterBandIdx = outputRasterBands[ outputRasterBandsIdx ];
        const unsigned int nRows = waveletRaster.getNumberOfRows();
        const unsigned int nCols = waveletRaster.getNumberOfColumns();        
        const unsigned int firstSmoothBandIdx = outputRasterBandsIdx *
          levelsNumber * 2;
        const unsigned int firstWaveletBandIdx = firstSmoothBandIdx
          + 1;          
        const unsigned int lastSmoothBandIdx = firstSmoothBandIdx
          + ( 2 * levelsNumber ) - 2;
        const unsigned int lastWaveletBandIdx = lastSmoothBandIdx + 1;
        unsigned int col = 0;
        unsigned int row = 0;        
        double value = 0;
        double sum = 0;
        unsigned int waveletRasterBand = 0;
        
        double bandAllowedMinValue = 0;
        double bandAllowedMaxValue = 0;
        te::rst::GetDataTypeRanges( outputRaster.getBand( 
          outputRasterBandIdx )->getProperty()->m_type, bandAllowedMinValue,
          bandAllowedMaxValue );
        
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            sum = 0.0;
            
            for( waveletRasterBand = firstWaveletBandIdx ; 
              waveletRasterBand <= lastWaveletBandIdx ; 
              waveletRasterBand += 2)
            {
              waveletRaster.getValue( col, row, value, waveletRasterBand );
              sum += value;
            }
            
            waveletRaster.getValue( col, row, value, lastSmoothBandIdx );
            sum += value;
            
            sum = std::max( sum, bandAllowedMinValue );
            sum = std::min( sum, bandAllowedMaxValue );
            
            outputRaster.setValue( col, row, sum, outputRasterBandIdx );
          }
        }
      }
           
      return true;
    }
    
    bool RasterResample( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      const te::rst::Interpolator::Method interpMethod, 
      const unsigned int firstRow,
      const unsigned int firstColumn, 
      const unsigned int height, 
      const unsigned int width,
      const unsigned int newheight, 
      const unsigned int newwidth, 
      const std::map<std::string, std::string>& rinfo,
      const std::string& dataSourceType,
      std::unique_ptr< te::rst::Raster >& resampledRasterPtr )
    {
      if( ( firstRow + height ) > inputRaster.getNumberOfRows() )
      {
        return false;
      }
      if( ( firstColumn + width ) > inputRaster.getNumberOfColumns() )
      {
        return false;
      }
      for (std::size_t inputRasterBandsIdx = 0; inputRasterBandsIdx < 
        inputRasterBands.size(); inputRasterBandsIdx++)
      {
        if( inputRasterBands[ inputRasterBandsIdx ] >= inputRaster.getNumberOfBands() )
        {
          return false;
        }
      }
      
      te::gm::Coord2D ulc = inputRaster.getGrid()->gridToGeo( (static_cast<double>(firstColumn))
        - 0.5, (static_cast<double>(firstRow)) - 0.5);
      te::gm::Coord2D lrc = inputRaster.getGrid()->gridToGeo( (static_cast<double>(firstColumn
        + width)) - 0.5, (static_cast<double>(firstRow + height)) - 0.5);
      
      std::unique_ptr< te::gm::Envelope > newEnvelopePtr( new te::gm::Envelope( 
        ulc.x, lrc.y, lrc.x, ulc.y ) );  

      // create output parameters and raster
      
      std::unique_ptr< te::rst::Grid > gridPtr( new te::rst::Grid(
        newwidth, newheight, newEnvelopePtr.release(),
        inputRaster.getSRID()) );

      std::vector<te::rst::BandProperty*> bands;

      for (std::size_t inputRasterBandsIdx = 0; inputRasterBandsIdx < 
        inputRasterBands.size(); inputRasterBandsIdx++)
      {
        bands.push_back( new te::rst::BandProperty(
          *(inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] )->getProperty())));
        bands[ inputRasterBandsIdx ]->m_blkh = 1;
        bands[ inputRasterBandsIdx ]->m_blkw = static_cast<int>(newwidth);
        bands[ inputRasterBandsIdx ]->m_nblocksx = 1;
        bands[ inputRasterBandsIdx ]->m_nblocksy = static_cast<int>(newheight);
      }

      resampledRasterPtr.reset( te::rst::RasterFactory::make( dataSourceType, 
        gridPtr.release(), bands, rinfo, 0 ) );
      if( resampledRasterPtr.get() == 0 )
      {
        return false;
      }

      // fill output raster
            
      std::complex<double> interpValue;
      te::rst::Interpolator interp( &inputRaster, interpMethod);      
      const double rowsFactor = (static_cast<double>(height-1)) / (static_cast<double>(newheight-1));
      const double colsFactor = (static_cast<double>(width-1)) / (static_cast<double>(newwidth-1));
      double inputRow = 0;
      double inputCol = 0;      
      unsigned int outputCol = 0;
      unsigned int outputRow = 0;
      unsigned int inputBandIdx = 0;
      double allowedMin = 0;
      double allowedMax = 0;
      
      for (std::size_t inputRasterBandsIdx = 0; inputRasterBandsIdx < 
        inputRasterBands.size(); inputRasterBandsIdx++)
      {
        te::rst::Band& outputBand = *resampledRasterPtr->getBand( inputRasterBandsIdx );
        inputBandIdx = inputRasterBands[ inputRasterBandsIdx ];
        te::rp::GetDataTypeRange( outputBand.getProperty()->m_type, allowedMin, 
          allowedMax );
        
        for ( outputRow = 0; outputRow < newheight; ++outputRow)
        {
          inputRow = ( (static_cast<double>( outputRow ) ) * rowsFactor ) + (static_cast<double>(firstRow));
          
          for ( outputCol = 0; outputCol < newwidth; ++outputCol )
          {
            inputCol = ( (static_cast<double>( outputCol ) ) * colsFactor ) + (static_cast<double>(firstColumn));
            
            interp.getValue(inputCol, inputRow, interpValue, inputBandIdx);
            
            interpValue.real( std::max( allowedMin, interpValue.real() ) );
            interpValue.real( std::min( allowedMax, interpValue.real() ) );

            outputBand.setValue(outputCol, outputRow, interpValue);
          }
        }
      }
      
      return true;
    }
    
    void CreateFixedStepPalette( 
      const unsigned int paletteSize, const bool randomize,
      std::vector< te::rst::BandProperty::ColorEntry >& palette )
    {
      palette.resize( paletteSize );
      if( paletteSize == 0 ) return;

      const short int step = static_cast<short int>( 256.0 /
        std::cbrt( static_cast<double>(paletteSize) ) );
      short int green = 0;
      short int blue = 0;
      unsigned int paletteIdx = 0;

      for( short int red = 0 ; red < 256 ; red += step )
      {
        for( green = 0 ; green < 256 ; green += step )
        {
          for( blue = 0 ; blue < 256 ; blue += step )
          {
            if( paletteIdx >= paletteSize )
            {
              green = 255;
              red = 255;
              break;
            }

            palette[ paletteIdx ].c1 = red;
            palette[ paletteIdx ].c2 = green;
            palette[ paletteIdx ].c3 = blue;
            palette[ paletteIdx ].c4 = 255;

            ++paletteIdx;
          }
        }
      }

      if( randomize )
      {
        std::random_shuffle( palette.begin(), palette.end() );
      }
    }
    
    bool RasterSlicing( 
      const te::rst::Raster& inputRaster,
      const unsigned int inputRasterBand,
      const bool createPaletteRaster,
      const unsigned int slicesNumber,
      const bool eqHistogram,
      const std::map< std::string, std::string >& rasterInfo,
      const std::string& rasterType,
      const bool enableProgress,
      std::vector< te::rst::BandProperty::ColorEntry > const * const palettePtr,
      std::unique_ptr< te::rst::Raster >& outRasterPtr ,
      std::map< double, double > limits)
    {
      TERP_TRUE_OR_RETURN_FALSE( inputRasterBand < inputRaster.getNumberOfBands(),
        "Invalid raster band index" );
      TERP_TRUE_OR_RETURN_FALSE( palettePtr != 0,
        "Invalid palette parameters" );
      
      te::rst::Raster const * internalInRasterPtr = &inputRaster;
      unsigned int internalInRasterBand = inputRasterBand;
      std::unique_ptr< te::rst::Raster > eqRasterPtr;
      if( eqHistogram )
      {
        double min = 0;
        double max = 0;
        GetDataTypeRange( inputRaster.getBand( inputRasterBand )->getProperty()->m_type,
          min, max );
        
        Contrast::InputParameters contInputPars;
        contInputPars.m_type = Contrast::InputParameters::HistogramEqualizationContrastT;
        contInputPars.m_inRasterPtr = &inputRaster;
        contInputPars.m_inRasterBands.push_back( inputRasterBand );
        contInputPars.m_hECMaxInput.push_back( max );
        contInputPars.m_enableProgress = enableProgress;
        
        Contrast contInstance;
        TERP_TRUE_OR_RETURN_FALSE( contInstance.initialize( contInputPars ),
          "Contrast init error" );        
        
        Contrast::OutputParameters contOutPars;
        contOutPars.m_createdOutRasterDSType = "EXPANSIBLE";
        TERP_TRUE_OR_RETURN_FALSE( contInstance.execute( contOutPars ),
          "Contrast exec error" );  
        
        eqRasterPtr.reset( contOutPars.m_createdOutRasterPtr.release() );
        internalInRasterPtr = eqRasterPtr.get();
        internalInRasterBand = 0;
      }

      std::vector< te::rst::BandProperty::ColorEntry > const * internalPalettePtr =
        palettePtr;
      std::vector< te::rst::BandProperty::ColorEntry > internalPalette;
      if( palettePtr == 0 )
      {
        CreateFixedStepPalette( slicesNumber, true, internalPalette );
        internalPalettePtr = &internalPalette;
      }
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setMessage( "Slicing" );
        progressPtr->setTotalSteps( static_cast<int>(2 * internalInRasterPtr->getNumberOfRows()) );
      }       

      if( createPaletteRaster )
      {
        std::vector< te::rst::BandProperty* > bandsProperties;
        bandsProperties.push_back( new te::rst::BandProperty( 
          * internalInRasterPtr->getBand( internalInRasterBand )->getProperty() ) );
        bandsProperties[ 0 ]->m_colorInterp = te::rst::PaletteIdxCInt;
        bandsProperties[ 0 ]->m_paletteInterp = te::rst::RGBPalInt;
        bandsProperties[ 0 ]->m_palette = (*internalPalettePtr);
        bandsProperties[ 0 ]->m_noDataValue = static_cast<double>(internalPalettePtr->size());
        if( internalPalettePtr->size() <= std::numeric_limits< unsigned char >::max() )
        {
          bandsProperties[ 0 ]->m_type = te::dt::UCHAR_TYPE;
        }
        else if( internalPalettePtr->size() <= std::numeric_limits< unsigned short int >::max() )
        {
          bandsProperties[ 0 ]->m_type = te::dt::UINT16_TYPE;
        }
        else if( internalPalettePtr->size() <= std::numeric_limits< unsigned int >::max() )
        {
          bandsProperties[ 0 ]->m_type = te::dt::UINT32_TYPE;
        }
        else
        {
          TERP_LOG_AND_RETURN_FALSE( "Invalid palette size" );
        }
        
        TERP_TRUE_OR_RETURN_FALSE( CreateNewRaster( *internalInRasterPtr->getGrid(),
          bandsProperties, rasterInfo, rasterType, outRasterPtr ),
          "Output raster creation error" );         
        
        const te::rst::Band& inBand = (*internalInRasterPtr->getBand( internalInRasterBand ));
        te::rst::Band& outBand = (*outRasterPtr->getBand( 0 ));
        const unsigned int nRows = internalInRasterPtr->getNumberOfRows();
        const unsigned int nCols = internalInRasterPtr->getNumberOfColumns();
         const double inputNoDataValue = inBand.getProperty()->m_noDataValue;
        const double outputNoDataValue = outBand.getProperty()->m_noDataValue;
        unsigned int col = 0;
        double value = 0;
        
        for( unsigned int row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, value );
            
            if( value == inputNoDataValue )
            {
              outBand.setValue( col, row, outputNoDataValue );
            }
            else
            {
              size_t index = 0;
              for (std::map<double, double>::iterator it = limits.begin(); it != limits.end(); ++it, ++index)
              {
                if (value >= (*it).first && value < (*it).second)
                {
                  outBand.setValue(col, row, (double)index);
                  break;
                }
              }
              if (index == limits.size())
                outBand.setValue(col, row, outputNoDataValue);
            }
          }
          
          if( enableProgress )
          {
            progressPtr->pulse();
            if( ! progressPtr->isActive() ) return false;
          }   
        }
      }
      else
      {
        short internalPaletteMaxValue = 0;
        for( unsigned int internalPaletteIdx = 0 ; internalPaletteIdx < 
          internalPalettePtr->size() ; ++internalPaletteIdx )
        {
          if( internalPalettePtr->operator[]( internalPaletteIdx ).c1 > 
            internalPaletteMaxValue )
          {
            internalPaletteMaxValue = internalPalettePtr->operator[]( internalPaletteIdx ).c1;
          }
          if( internalPalettePtr->operator[]( internalPaletteIdx ).c2 > 
            internalPaletteMaxValue )
          {
            internalPaletteMaxValue = internalPalettePtr->operator[]( internalPaletteIdx ).c2;
          }
          if( internalPalettePtr->operator[]( internalPaletteIdx ).c3 > 
            internalPaletteMaxValue )
          {
            internalPaletteMaxValue = internalPalettePtr->operator[]( internalPaletteIdx ).c3;
          }
          if( internalPalettePtr->operator[]( internalPaletteIdx ).c4 > 
            internalPaletteMaxValue )
          {
            internalPaletteMaxValue = internalPalettePtr->operator[]( internalPaletteIdx ).c4;
          }                              
        }    
        
        int outDataType = 0;
        if( internalPaletteMaxValue < ( std::numeric_limits< unsigned char >::max() ) )
        {
          outDataType = te::dt::UCHAR_TYPE;
        }
        else if( internalPaletteMaxValue < ( std::numeric_limits< unsigned short int >::max() ) )
        {
          outDataType = te::dt::UINT16_TYPE;
        }
        else
        {
          outDataType = te::dt::UINT32_TYPE;
        }         
        
        std::vector< te::rst::BandProperty* > bandsProperties;
        bandsProperties.push_back( new te::rst::BandProperty( 
          * internalInRasterPtr->getBand( internalInRasterBand )->getProperty() ) );
        bandsProperties.push_back( new te::rst::BandProperty( 
          * internalInRasterPtr->getBand( internalInRasterBand )->getProperty() ) );
        bandsProperties.push_back( new te::rst::BandProperty( 
          * internalInRasterPtr->getBand( internalInRasterBand )->getProperty() ) );        
        bandsProperties[ 0 ]->m_colorInterp = te::rst::RedCInt;
        bandsProperties[ 1 ]->m_colorInterp = te::rst::GreenCInt;
        bandsProperties[ 2 ]->m_colorInterp = te::rst::BlueCInt;
        bandsProperties[ 0 ]->m_type = outDataType;
        bandsProperties[ 1 ]->m_type = outDataType;
        bandsProperties[ 2 ]->m_type = outDataType;
        bandsProperties[ 0 ]->m_noDataValue = internalPaletteMaxValue + 1;
        bandsProperties[ 1 ]->m_noDataValue = internalPaletteMaxValue + 1;
        bandsProperties[ 2 ]->m_noDataValue = internalPaletteMaxValue + 1;
        
        TERP_TRUE_OR_RETURN_FALSE( CreateNewRaster( *internalInRasterPtr->getGrid(),
          bandsProperties, rasterInfo, rasterType, outRasterPtr ),
          "Output raster creation error" );         
        
        const te::rst::Band& inBand = (*internalInRasterPtr->getBand( internalInRasterBand ));
        te::rst::Band& outBandRed = (*outRasterPtr->getBand( 0 ));
        te::rst::Band& outBandGreen = (*outRasterPtr->getBand( 1 ));
        te::rst::Band& outBandBlue = (*outRasterPtr->getBand( 2 ));
        const double rBandNoDataValue = outBandRed.getProperty()->m_noDataValue;
        const double gBandNoDataValue = outBandGreen.getProperty()->m_noDataValue;
        const double bBandNoDataValue = outBandBlue.getProperty()->m_noDataValue;
        const unsigned int nRows = internalInRasterPtr->getNumberOfRows();
        const unsigned int nCols = internalInRasterPtr->getNumberOfColumns();
        const double inputNoDataValue = inBand.getProperty()->m_noDataValue;
        unsigned int col = 0;
        double value = 0;
        
        for( unsigned int row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, value );
            
            if( value == inputNoDataValue )
            {
              outBandRed.setValue( col, row, rBandNoDataValue );
              outBandGreen.setValue( col, row, gBandNoDataValue );
              outBandBlue.setValue( col, row, bBandNoDataValue );
            }
            else
            {
                size_t index = 0;
                for (std::map<double, double>::iterator it = limits.begin(); it != limits.end(); ++it, ++index)
                {
                  if (value >= (*it).first && value < (*it).second)
                    break;
                }

                if (index < limits.size())
                {
                  const te::rst::BandProperty::ColorEntry& cEntry = internalPalettePtr->operator[](static_cast<unsigned int>(index));

                  outBandRed.setValue(col, row, cEntry.c1);
                  outBandGreen.setValue(col, row, cEntry.c2);
                  outBandBlue.setValue(col, row, cEntry.c3);
                }
                else
                {
                  outBandRed.setValue(col, row, rBandNoDataValue);
                  outBandGreen.setValue(col, row, gBandNoDataValue);
                  outBandBlue.setValue(col, row, bBandNoDataValue);
                }
            }
          }
          
          if( enableProgress )
          {
            progressPtr->pulse();
            if( ! progressPtr->isActive() ) return false;
          }   
        } 
      }
     
      return true;
    }
    
  } // end namespace rp
}   // end namespace te
